# =====================================
#   perf_epoch_bridge.py
# =====================================
"""
Bi-directional conversion utilities between UNIX-epoch nanoseconds
and the Python perf_counter() clock.

Usage:
    >>> import perf_epoch_bridge as peb
    >>> peb.init_clock_bridge()
    >>> span_start = peb.epoch_nanos_to_perf_seconds(start_time_unix_nano)
    >>> span_end   = peb.epoch_nanos_to_perf_seconds(end_time_unix_nano)
    >>> duration   = span_end - span_start
"""

from __future__ import annotations
import time
from typing import Final, Union

# Module globals are initialised exactly once.
_anchor_perf_ns: Union[int, None] = None
_anchor_wall_ns: Union[int, None] = None
_offset_ns: Union[int, None] = None


def init_clock_bridge() -> None:
    """Capture simultaneous perf & wall-clock samples and compute offset."""
    global _anchor_perf_ns, _anchor_wall_ns, _offset_ns

    # Capture as closely together as possible
    _anchor_perf_ns = time.perf_counter_ns()
    _anchor_wall_ns = time.time_ns()
    _offset_ns = _anchor_perf_ns - _anchor_wall_ns


def epoch_nanos_to_perf_seconds(epoch_ns: int) -> float:
    """Translate a UNIX epoch (ns) timestamp onto perf_counter() seconds."""
    if _offset_ns is None:
        raise RuntimeError("init_clock_bridge() must be called first!")
    return (epoch_ns + _offset_ns) / 1_000_000_000.0


def perf_seconds_now() -> float:
    """Return current perf_counter() reading (seconds)."""
    return time.perf_counter()


# Optional: reverse conversion (perf → epoch)
def perf_seconds_to_epoch_nanos(perf_sec: float) -> int:
    """Translate a perf_counter() float back to epoch nanoseconds."""
    if _offset_ns is None:
        raise RuntimeError("init_clock_bridge() must be called first!")
    return int(perf_sec * 1_000_000_000) - _offset_ns



from typing import Optional, List
import textwrap


class TurnFaithfulnessTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
        - When evaluating claims, compare them to BOTH textual and visual evidence.
        - If the claim references something not clearly visible, respond with 'idk'.
    """

    @staticmethod
    def generate_claims(
        input: str, assistant_output: str, multimodal: bool = False
    ):
        return textwrap.dedent(
            f"""
            Extract every factual-sounding claim asserted in the ASSISTANT'S OUTPUT.

            A claim is any statement presented as fact, even if it is incorrect, vague, implied, or unverifiable.

            RULES:
            - Use ONLY the assistant's output as the source of claims.
            - Use the user's preceding message ONLY to resolve pronouns or references, not as factual evidence.
            - Extract claims exactly as stated without rewriting, summarizing, merging, or omitting details.
            - If a sentence contains multiple factual assertions, extract each as a separate claim.
            - Claims may involve text or images if multimodal.
            - Do NOT add, infer, or transform information.

            {TurnFaithfulnessTemplate.multimodal_rules if multimodal else ""}

            Output MUST be ONLY valid JSON:

            {{
                "claims": ["claim 1", "claim 2", ...]
            }}

            USER MESSAGE:
            {input}

            ASSISTANT OUTPUT:
            {assistant_output}

            JSON:
            """
        )

    @staticmethod
    def generate_truths(
        reference_context: str,
        extraction_limit: Optional[int],
        multimodal: bool = False,
    ):
        if extraction_limit is None:
            limit_description = "factual, explicit truths"
        elif extraction_limit == 1:
            limit_description = "one factual, explicit truth"
        else:
            limit_description = f"{extraction_limit} factual, explicit truths"

        return textwrap.dedent(
            f"""
            Extract {limit_description} from the REFERENCE CONTEXT.

            RULES:
            - Truths must be atomic, explicit factual statements.
            - Do not summarize or combine multiple facts.
            - Select truths based on reading order, not 'importance'.
            - Do not infer or expand beyond what is explicitly stated.
            - Keep each truth minimal but complete.
            - Treat images as factual evidence if multimodal, using only clearly visible information.

            {TurnFaithfulnessTemplate.multimodal_rules if multimodal else ""}

            Output MUST be ONLY valid JSON:

            {{
                "truths": ["truth 1", "truth 2", ...]
            }}

            REFERENCE CONTEXT:
            {reference_context}

            JSON:
            """
        )

    @staticmethod
    def generate_verdicts(
        claims: List[str], reference_context: str, multimodal: bool = False
    ):
        return textwrap.dedent(
            f"""
            For each claim, determine whether it is supported, contradicted, or not addressed by the reference context.

            DEFINITIONS:
            - "yes"  = The claim is directly supported by at least one truth.
            - "no"   = The claim directly contradicts at least one truth.
            - "idk"  = The context does not confirm or contradict the claim.

            RULES:
            - One verdict per claim, in the same order.
            - Do NOT use prior knowledge.
            - Only use the explicit truths provided.
            - A "yes" verdict must not include a reason.
            - A "no" or "idk" verdict must include a concise reason that quotes or paraphrases only the truths.
            - If a claim references an image and the visibility is unclear or ambiguous, use "idk".
            - Do not create new facts or explanations.

            {TurnFaithfulnessTemplate.multimodal_rules if multimodal else ""}

            Output MUST be ONLY valid JSON:

            {{
                "verdicts": [
                    {{
                        "verdict": "yes"
                    }},
                    {{
                        "verdict": "no",
                        "reason": "<explanation>"
                    }},
                    {{
                        "verdict": "idk",
                        "reason": "<explanation>"
                    }}
                ]
            }}

            REFERENCE CONTEXT:
            {reference_context}

            CLAIMS:
            {claims}

            JSON:
            """
        )

    @staticmethod
    def generate_reason(
        score: float, contradictions: List[str], multimodal: bool = False
    ):
        return textwrap.dedent(
            f"""
            Below is a list of contradictions extracted from verdicts. Write a concise justification of the score.

            RULES:
            - If contradictions exist, summarize them in 1-3 sentences.
            - If no contradictions exist, respond:
              {{
                  "reason": "No contradictions were found."
              }}
            - The summary must reference only the contradictions listed.
            - Tone must be neutral and concise.
            - No external knowledge may be used.

            {TurnFaithfulnessTemplate.multimodal_rules if multimodal else ""}

            Output MUST be ONLY valid JSON:

            {{
                "reason": "<summary>"
            }}

            FAITHFULNESS SCORE:
            {score}

            CONTRADICTIONS:
            {contradictions}

            JSON:
            """
        )

    @staticmethod
    def generate_final_reason(
        final_score: float, success: bool, reasons: List[str]
    ):
        return textwrap.dedent(
            f"""You are an AI evaluator producing a single final explanation for the TurnFaithfulnessMetric result.

                Context:
                This metric evaluates conversational faithfulness by extracting truths from retrieval context, extracting claims from the assistant's output, and generating verdicts that compare each claim against the truths. Each interaction yields a reason indicating why a verdict failed or succeeded. You are given all those reasons.

                Inputs:
                - final_score: the averaged score across all interactions.
                - success: whether the metric passed or failed
                - reasons: a list of textual reasons generated from individual verdicts.

                Instructions:
                1. Read all reasons and synthesize them into one unified explanation.
                2. Describe patterns of claim-truth mismatches, contradictions, hallucinations, unsupported statements, or image-related errors if present.
                3. Do not repeat every reason; merge them into a concise, coherent narrative.
                5. If the metric failed, state the dominant failure modes. If it passed, state why the model's claims aligned with truths.
                6. Output a single paragraph with no lists, no bullets, no markup.

                Output:
                A single paragraph explaining the final outcome.

                Here's the inputs:

                Final Score: {final_score}
                
                Reasons: 
                {reasons}

                Success: {success}

                Now give me a final reason that explains why the metric passed or failed. Output ONLY the reason and nothing else.

                The final reason:
            """
        )



class ExtractionTemplate:

    @staticmethod
    def extract_prompt_structure_from_inputs(inputs: list[str]):
        return f"""
            You are a prompt engineer tasked with reverse-engineering the original prompt that may have produced the following inputs. 
            Each input is a message that a user might submit to an AI system.

            Your job is to infer the structure of the original prompt by analyzing patterns in these inputs.

            Specifically, extract the following:
            
            1. `scenario`: Describe the type of person or user who would have submitted these inputs, and the context or purpose for doing so.
            2. `task`: What was the AI system expected to do in response to these inputs?
            3. `input_format`: Describe the style, tone, or structure of the inputs — how the inputs are typically phrased.

            You MUST return your answer strictly in the following JSON format:

            ```json
            {{
                "scenario": "<your answer here>",
                "task": "<your answer here>",
                "input_format": "<your answer here>"
            }}
            ```

            **
            IMPORTANT: Do not use any prior knowledge. Only rely on what is observable in the inputs themselves.   

            Example inputs: [
                "How many users signed up last week?",
                "Show me the total revenue for March.",
                "Which products had the highest sales yesterday?"
            ]

            Example output:
            {{
                "scenario": "Non-technical users trying to query a database using plain English.",
                "task": "Answering text-to-SQL-related queries by querying a database and returning the results to users.",
                "input_format": "Questions in English that ask for data in a database."
            }}

            Here are the inputs to analyze:

            {inputs}
     """

    @staticmethod
    def extract_conversational_structure_from_scenarios(example_scenarios):
        scenarios_text = "\n".join(
            [f"- {scenario}" for scenario in example_scenarios]
        )

        return f"""Analyze the following conversational scenarios and extract the common structural elements:

        Example Scenarios:
        {scenarios_text}

        Based on these examples, identify and return in JSON format:
        1. **scenario_context**: The general context or domain in which these conversations occur (e.g., "customer service", "educational settings", "workplace interactions")
        2. **conversational_task**: The primary goal or purpose these conversations aim to achieve (e.g., "resolve issues", "provide information", "give feedback")
        3. **participant_roles**: The typical participants involved in these conversations (e.g., "customer and support agent", "teacher and student", "manager and employee")

        **
        IMPORTANT: Please make sure to only return in JSON format, with the 'scenario_context', 'conversational_task', and 'participant_roles' keys.

        Example JSON:
        {{
            "scenario_context": "Educational settings and academic discussions",
            "conversational_task": "Explain concepts and answer questions",
            "participant_roles": "Teacher and student, or peer students"
        }}

        The values MUST be STRINGS that capture the essence of the conversational patterns in the examples.
        **

        JSON:
        """



import inspect
from typing import Callable, Union

from deepeval.metrics import BaseConversationalMetric, BaseMetric
from deepeval.test_case import ConversationalTestCase, LLMTestCase


def _build_measure_kwargs(func: Callable) -> dict:
    params = inspect.signature(func).parameters
    kwargs = {}
    for key in ("_show_indicator", "_in_component", "_log_metric_to_confident"):
        if key in params:
            kwargs[key] = False
    return kwargs


def _measure_no_indicator(
    metric: Union[BaseMetric, BaseConversationalMetric],
    test_case: Union[LLMTestCase, ConversationalTestCase],
):
    kwargs = _build_measure_kwargs(metric.measure)
    return metric.measure(test_case, **kwargs)


async def _a_measure_no_indicator(
    metric: Union[BaseMetric, BaseConversationalMetric],
    test_case: Union[LLMTestCase, ConversationalTestCase],
):
    kwargs = _build_measure_kwargs(metric.a_measure)
    return await metric.a_measure(test_case, **kwargs)



from typing import List, Optional, Literal
from pydantic import BaseModel, Field


class Misuses(BaseModel):
    misuses: List[str]


class MisuseVerdict(BaseModel):
    verdict: Literal["yes", "no"]
    reason: Optional[str] = Field(default=None)


class Verdicts(BaseModel):
    verdicts: List[MisuseVerdict]


class MisuseScoreReason(BaseModel):
    reason: str



from typing import Union, Dict, Optional, List
from deepeval.test_run import global_test_run_manager
from deepeval.prompt import Prompt
from deepeval.prompt.api import PromptApi
from deepeval.test_run.test_run import TEMP_FILE_PATH
from deepeval.confident.api import is_confident
from deepeval.test_run.test_run import PromptData


def process_hyperparameters(
    hyperparameters: Optional[Dict] = None,
    verbose: bool = True,
) -> Union[Dict[str, Union[str, int, float, PromptApi]], None]:
    if hyperparameters is None:
        return None

    if not isinstance(hyperparameters, dict):
        raise TypeError("Hyperparameters must be a dictionary or None")

    processed_hyperparameters = {}
    prompts_version_id_map = {}

    for key, value in hyperparameters.items():
        if not isinstance(key, str):
            raise TypeError(f"Hyperparameter key '{key}' must be a string")

        if value is None:
            continue

        if not isinstance(value, (str, int, float, Prompt)):
            raise TypeError(
                f"Hyperparameter value for key '{key}' must be a string, integer, float, or Prompt"
            )

        if isinstance(value, Prompt):
            try:
                prompt_key = f"{value.alias}_{value.version}"
            except AttributeError:
                prompt_key = f"{value.alias}_00.00.01"

            if value._prompt_version_id is not None and value.type is not None:
                processed_hyperparameters[key] = PromptApi(
                    id=value._prompt_version_id,
                    type=value.type,
                )
            elif is_confident():
                if prompt_key not in prompts_version_id_map:
                    value.push(_verbose=verbose)
                    prompts_version_id_map[prompt_key] = (
                        value._prompt_version_id
                    )
                processed_hyperparameters[key] = PromptApi(
                    id=prompts_version_id_map[prompt_key],
                    type=value.type,
                )
        else:
            processed_hyperparameters[key] = str(value)

    return processed_hyperparameters


def log_hyperparameters(func):
    test_run = global_test_run_manager.get_test_run()

    def modified_hyperparameters():
        base_hyperparameters = func()
        return base_hyperparameters

    hyperparameters = process_hyperparameters(modified_hyperparameters())
    test_run.hyperparameters = hyperparameters
    global_test_run_manager.save_test_run(TEMP_FILE_PATH)

    # Define the wrapper function that will be the actual decorator
    def wrapper(*args, **kwargs):
        # Optional: You can decide if you want to do something else here
        # every time the decorated function is called
        return func(*args, **kwargs)

    # Return the wrapper function to be used as the decorator
    return wrapper


def process_prompts(
    hyperparameters: Dict[str, Union[str, int, float, Prompt]],
) -> List[PromptData]:
    prompts = []
    if not hyperparameters:
        return prompts
    seen_prompts = set()
    prompt_objects = [
        value for value in hyperparameters.values() if isinstance(value, Prompt)
    ]
    for prompt in prompt_objects:
        prompt_version = prompt.version if is_confident() else None
        prompt_key = f"{prompt.alias}_{prompt_version}"
        if prompt_key in seen_prompts:
            continue
        seen_prompts.add(prompt_key)
        prompt_data = PromptData(
            alias=prompt.alias,
            version=prompt_version,
            text_template=prompt.text_template,
            messages_template=prompt.messages_template,
            model_settings=prompt.model_settings,
            output_type=prompt.output_type,
            interpolation_type=prompt.interpolation_type,
        )
        prompts.append(prompt_data)
    return prompts



from typing import Literal, Dict, Union
from pydantic import BaseModel


class MetricScoreReason(BaseModel):
    reason: str


class TaskNodeOutput(BaseModel):
    output: Union[str, list[str], dict[str, str]]


class BinaryJudgementVerdict(BaseModel):
    verdict: bool
    reason: str


class NonBinaryJudgementVerdict(BaseModel):
    verdict: str
    reason: str



from typing import List, Union
import textwrap
from deepeval.test_case import MLLMImage


class TurnContextualRelevancyTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
        - When evaluating claims, compare them to BOTH textual and visual evidence.
        - If the claim references something not clearly visible, respond with 'idk'.
    """

    @staticmethod
    def generate_reason(
        input: Union[str, List[Union[str, MLLMImage]]],
        irrelevant_statements: List[str],
        relevant_statements: List[str],
        score: float,
        multimodal: bool = False,
    ):
        # Note: irrelevancies parameter name in multimodal version is kept as irrelevant_statements for consistency
        return textwrap.dedent(
            f"""Based on the given user message, reasons for why the retrieval context is irrelevant to the user message, the statements in the retrieval context that is actually relevant to the retrieval context, and the contextual relevancy score (the closer to 1 the better), please generate a CONCISE reason for the score.
            In your reason, you should quote data provided in the reasons for irrelevancy and relevant statements to support your point.

            {TurnContextualRelevancyTemplate.multimodal_rules if multimodal else ""}
            
            ** 
            IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
            Example JSON:
            {{
                "reason": "The score is <contextual_relevancy_score> because <your_reason>."
            }}

            If the score is 1, keep it short and say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying).
            **

            Contextual Relevancy Score:
            {score}

            User Message:
            {input}
            
            Reasons for why the retrieval context is irrelevant to the user message:
            {irrelevant_statements}

            Statement in the retrieval context that is relevant to the user message:
            {relevant_statements}

            JSON:
            """
        )

    @staticmethod
    def generate_verdicts(
        input: Union[str, List[Union[str, MLLMImage]]],
        context: Union[str, List[Union[str, MLLMImage]]],
        multimodal: bool = False,
    ):
        context_type = "context (image or string)" if multimodal else "context"
        statement_or_image = "statement or image" if multimodal else "statement"

        # Conditional instructions based on mode
        extraction_instructions = ""
        if multimodal:
            extraction_instructions = textwrap.dedent(
                """
                If the context is textual, you should first extract the statements found in the context if the context, which are high level information found in the context, before deciding on a verdict and optionally a reason for each statement.
                If the context is an image, `statement` should be a description of the image. Do not assume any information not visibly available.
                """
            ).strip()
        else:
            extraction_instructions = "You should first extract statements found in the context, which are high level information found in the context, before deciding on a verdict and optionally a reason for each statement."

        # Additional instruction for empty context (only in non-multimodal)
        empty_context_instruction = ""
        if not multimodal:
            empty_context_instruction = '\nIf provided context contains no actual content or statements then: give "no" as a "verdict",\nput context into "statement", and "No statements found in provided context." into "reason".'

        return textwrap.dedent(
            f"""Based on the user message and {context_type}, please generate a JSON object to indicate whether {'the context' if multimodal else 'each statement found in the context'} is relevant to the provided user message. The JSON will be a list of 'verdicts', with 2 mandatory fields: 'verdict' and 'statement', and 1 optional field: 'reason'.
            {extraction_instructions}
            The 'verdict' key should STRICTLY be either 'yes' or 'no', and states whether the {statement_or_image} is relevant to the user message.
            Provide a 'reason' ONLY IF verdict is no. You MUST quote the irrelevant parts of the {statement_or_image} to back up your reason.{empty_context_instruction}
            
            {TurnContextualRelevancyTemplate.multimodal_rules if multimodal else ""}

            **
            IMPORTANT: Please make sure to only return in JSON format.
            Example Context: "Einstein won the Nobel Prize for his discovery of the photoelectric effect. He won the Nobel Prize in 1968. There was a cat."
            Example User Message: "What were some of Einstein's achievements?"

            Example:
            {{
                "verdicts": [
                    {{
                        "statement": "Einstein won the Nobel Prize for his discovery of the photoelectric effect in 1968",
                        "verdict": "yes"
                    }},
                    {{
                        "statement": "There was a cat.",
                        "reason": "The retrieval context contained the information 'There was a cat' when it has nothing to do with Einstein's achievements.",
                        "verdict": "no"
                    }}
                ]
            }}
            **

            User Message:
            {input}

            Context:
            {context}

            JSON:
            """
        )

    @staticmethod
    def generate_final_reason(
        final_score: float, success: bool, reasons: List[str]
    ):
        return textwrap.dedent(
            f"""You are an AI evaluator producing a single final explanation for the TurnContextualRelevancyMetric result.

            Context:
            This metric evaluates conversational contextual relevancy by determining whether statements in the retrieval context are relevant to the user message for each interaction. Each interaction yields a reason indicating which statements were relevant or irrelevant. You are given all those reasons.

            Inputs:
            - final_score: the averaged score across all interactions.
            - success: whether the metric passed or failed
            - reasons: a list of textual reasons generated from individual interactions.

            Instructions:
            1. Read all reasons and synthesize them into one unified explanation.
            2. Describe patterns of irrelevant statements, off-topic context, or well-targeted retrieval if present.
            3. Do not repeat every reason; merge them into a concise, coherent narrative.
            4. If the metric failed, state the dominant failure modes. If it passed, state why the retrieval context was relevant to user messages.
            5. Output a single paragraph with no lists, no bullets, no markup.

            Output:
            A single paragraph explaining the final outcome.

            Here's the inputs:

            Final Score: {final_score}
            
            Reasons: 
            {reasons}

            Success: {success}

            Now give me a final reason that explains why the metric passed or failed. Output ONLY the reason and nothing else.

            The final reason:
            """
        )



from __future__ import annotations
import asyncio
import copy
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Union,
)

from deepeval.dataset.golden import Golden, ConversationalGolden
from deepeval.dataset.utils import (
    convert_goldens_to_test_cases,
    convert_convo_goldens_to_convo_test_cases,
)
from deepeval.errors import DeepEvalError
from deepeval.metrics import (
    BaseMetric,
    BaseConversationalMetric,
)
from deepeval.metrics.utils import copy_metrics
from deepeval.test_case import (
    LLMTestCase,
    ConversationalTestCase,
    Turn,
)
from deepeval.prompt.prompt import Prompt

from deepeval.optimizer.types import (
    ModelCallback,
    PromptConfiguration,
    Objective,
    MeanObjective,
    ModuleId,
)
from deepeval.optimizer.scorer.base import BaseScorer
from deepeval.optimizer.utils import (
    validate_callback,
    validate_metrics,
    invoke_model_callback,
    a_invoke_model_callback,
)
from deepeval.optimizer.scorer.utils import (
    _measure_no_indicator,
    _a_measure_no_indicator,
)


class Scorer(BaseScorer):
    """
    Scores prompts by running model_callback, building test cases,
    running metrics, and aggregating scores.
    """

    DEFAULT_MODULE_ID: ModuleId = "__module__"

    def __init__(
        self,
        model_callback: ModelCallback,
        metrics: Union[List[BaseMetric], List[BaseConversationalMetric]],
        max_concurrent: int,
        throttle_seconds: float,
        objective_scalar: Objective = MeanObjective(),
    ):
        self.model_callback = validate_callback(
            component="Scorer",
            model_callback=model_callback,
        )
        self.metrics = validate_metrics(component="Scorer", metrics=metrics)
        self.objective_scalar = objective_scalar
        self._semaphore = asyncio.Semaphore(max_concurrent)
        self._throttle = float(throttle_seconds)

    ########################
    # generation & scoring #
    ########################

    def generate(
        self,
        prompts_by_module: Dict[ModuleId, Prompt],
        golden: Union[Golden, ConversationalGolden],
    ) -> str:
        module_id = self._select_module_id_from_prompts(prompts_by_module)
        prompt = prompts_by_module.get(module_id) or next(
            iter(prompts_by_module.values())
        )

        return invoke_model_callback(
            model_callback=self.model_callback,
            prompt=prompt,
            golden=golden,
        )

    async def a_generate(
        self,
        prompts_by_module: Dict[ModuleId, Prompt],
        golden: Union[Golden, ConversationalGolden],
    ) -> str:
        module_id = self._select_module_id_from_prompts(prompts_by_module)
        prompt = prompts_by_module.get(module_id) or next(
            iter(prompts_by_module.values())
        )

        return await a_invoke_model_callback(
            model_callback=self.model_callback,
            prompt=prompt,
            golden=golden,
        )

    def score_pareto(
        self,
        prompt_configuration: PromptConfiguration,
        d_pareto: Union[List[Golden], List[ConversationalGolden]],
    ) -> List[float]:
        return [
            self._score_one(prompt_configuration, golden) for golden in d_pareto
        ]

    def score_minibatch(
        self,
        prompt_configuration: PromptConfiguration,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> float:
        if not minibatch:
            return 0.0

        scores = [
            self._score_one(prompt_configuration, golden)
            for golden in minibatch
        ]
        return sum(scores) / len(scores)

    def get_minibatch_feedback(
        self,
        prompt_configuration: PromptConfiguration,
        module: ModuleId,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> str:
        # default metric feedback (μ_f): concat metric.reason across minibatch and cap length
        reasons: List[str] = []
        for golden in minibatch:
            actual = self.generate(prompt_configuration.prompts, golden)
            test_case = self._golden_to_test_case(golden, actual)
            for metric in copy_metrics(self.metrics):
                _measure_no_indicator(metric=metric, test_case=test_case)
                if metric.reason:
                    reasons.append(str(metric.reason))
        if not reasons:
            return ""
        unique: List[str] = []
        seen = set()
        for reason in reasons:
            if reason not in seen:
                unique.append(reason)
                seen.add(reason)
        return "\n---\n".join(
            unique[:8]
        )  # TODO: Make how much feedback configurable

    async def a_score_pareto(
        self,
        prompt_configuration: PromptConfiguration,
        d_pareto: Union[List[Golden], List[ConversationalGolden]],
    ) -> List[float]:
        tasks = [
            self._bounded(self._a_score_one(prompt_configuration, golden))
            for golden in d_pareto
        ]
        return await asyncio.gather(*tasks)

    async def a_score_minibatch(
        self,
        prompt_configuration: PromptConfiguration,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> float:
        tasks = [
            self._bounded(self._a_score_one(prompt_configuration, golden))
            for golden in minibatch
        ]
        scores = await asyncio.gather(*tasks)
        return sum(scores) / len(scores) if scores else 0.0

    async def a_get_minibatch_feedback(
        self,
        prompt_configuration: PromptConfiguration,
        module: ModuleId,
        minibatch: Union[List[Golden], List[ConversationalGolden]],
    ) -> str:
        async def reasons_one(golden) -> List[str]:
            # Clone per task to avoid shared state
            metrics = copy_metrics(self.metrics)
            # metrics = self.metrics
            actual = await self.a_generate(prompt_configuration.prompts, golden)
            test_case = self._golden_to_test_case(golden, actual)
            out: List[str] = []
            for metric in metrics:
                await _a_measure_no_indicator(metric, test_case)
                if metric.reason:
                    out.append(str(metric.reason))
            return out

        tasks = [self._bounded(reasons_one(golden)) for golden in minibatch]
        nested = await asyncio.gather(*tasks)
        reasons: List[str] = [reason for sub in nested for reason in sub]
        if not reasons:
            return ""
        unique: List[str] = []
        seen = set()
        for reason in reasons:
            if reason not in seen:
                unique.append(reason)
                seen.add(reason)
        return "\n---\n".join(unique[:8])

    ###################
    # scoring helpers #
    ###################

    def _golden_to_test_case(
        self,
        golden: Union[Golden, ConversationalGolden],
        actual: str,
    ) -> Union[LLMTestCase, ConversationalTestCase]:
        """Convert a golden + actual output into a test case for metrics."""
        if isinstance(golden, Golden):
            golden.actual_output = actual
            return convert_goldens_to_test_cases([golden])[0]

        if isinstance(golden, ConversationalGolden):
            # Build turns with actual output as assistant response
            turns: List[Turn] = list(golden.turns or [])
            if turns and turns[-1].role == "assistant":
                turns[-1] = Turn(role="assistant", content=actual)
            elif turns:
                turns.append(Turn(role="assistant", content=actual))
            else:
                turns = [
                    Turn(role="assistant", content=actual),
                ]

            golden.turns = turns
            return convert_convo_goldens_to_convo_test_cases([golden])[0]

    async def _bounded(self, coro):
        if self._semaphore is None:
            return await coro
        async with self._semaphore:
            res = await coro
        if self._throttle:
            await asyncio.sleep(self._throttle)
        return res

    async def _a_score_one(
        self,
        prompt_configuration: PromptConfiguration,
        golden: Union[Golden, ConversationalGolden],
    ) -> float:
        # Clone metrics to avoid shared-state
        metrics = copy_metrics(self.metrics)
        actual = await self.a_generate(prompt_configuration.prompts, golden)
        test_case = self._golden_to_test_case(golden, actual)

        per_metric: Dict[str, float] = {}
        for metric in metrics:
            score = await _a_measure_no_indicator(metric, test_case)
            per_metric[metric.__class__.__name__] = float(score)
        return self.objective_scalar.scalarize(per_metric)

    def _score_one(
        self,
        prompt_configuration: PromptConfiguration,
        golden: Union[Golden, ConversationalGolden],
    ) -> float:
        metrics = copy_metrics(self.metrics)
        actual = self.generate(prompt_configuration.prompts, golden)
        test_case = self._golden_to_test_case(golden, actual)

        per_metric: Dict[str, float] = {}
        for metric in metrics:
            score = _measure_no_indicator(metric, test_case)
            per_metric[metric.__class__.__name__] = float(score)
        return self.objective_scalar.scalarize(per_metric)

    def _select_module_id_from_prompts(
        self, prompts_by_module: Dict[ModuleId, Prompt]
    ) -> ModuleId:
        """
        Default module selection strategy:

        - Prefer the synthetic '__module__' key when present
        - Otherwise fall back to the first key in prompts_by_module.

        Assumes `prompts_by_module` is non-empty; callers should validate that.
        """
        if self.DEFAULT_MODULE_ID in prompts_by_module:
            return self.DEFAULT_MODULE_ID

        # At this point we expect at least one key.
        try:
            return next(iter(prompts_by_module.keys()))
        except StopIteration:
            raise DeepEvalError(
                "Scorer._select_module_id_from_prompts(...) "
                "received an empty `prompts_by_module`. At least one Prompt is required."
            )

    def select_module(
        self, prompt_configuration: PromptConfiguration
    ) -> ModuleId:
        return self._select_module_id_from_prompts(prompt_configuration.prompts)

    async def a_select_module(
        self, prompt_configuration: PromptConfiguration
    ) -> ModuleId:
        return self.select_module(prompt_configuration)



from abc import ABC, abstractmethod
from typing import Any, Optional, List, Union
from deepeval.models.utils import parse_model_name
from dataclasses import dataclass


@dataclass
class DeepEvalModelData:
    supports_log_probs: Optional[bool] = None
    supports_multimodal: Optional[bool] = None
    supports_structured_outputs: Optional[bool] = None
    supports_json: Optional[bool] = None
    input_price: Optional[float] = None
    output_price: Optional[float] = None
    supports_temperature: Optional[bool] = True


class DeepEvalBaseModel(ABC):
    def __init__(self, model_name: Optional[str] = None, *args, **kwargs):
        self.model_name = model_name
        self.model = self.load_model(*args, **kwargs)

    @abstractmethod
    def load_model(self, *args, **kwargs) -> "DeepEvalBaseModel":
        """Loads a model, that will be responsible for scoring.

        Returns:
            A model object
        """
        pass

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._call(*args, **kwargs)

    @abstractmethod
    def _call(self, *args, **kwargs):
        """Runs the model to score / output the model predictions.

        Returns:
            A score or a list of results.
        """
        pass


class DeepEvalBaseLLM(ABC):
    def __init__(self, model: Optional[str] = None, *args, **kwargs):
        self.name = parse_model_name(model)
        self.model = self.load_model()

    @abstractmethod
    def load_model(self, *args, **kwargs) -> "DeepEvalBaseLLM":
        """Loads a model, that will be responsible for scoring.

        Returns:
            A model object
        """
        pass

    @abstractmethod
    def generate(self, *args, **kwargs) -> str:
        """Runs the model to output LLM response.

        Returns:
            A string.
        """
        pass

    @abstractmethod
    async def a_generate(self, *args, **kwargs) -> str:
        """Runs the model to output LLM response.

        Returns:
            A string.
        """
        pass

    @abstractmethod
    def get_model_name(self, *args, **kwargs) -> str:
        return self.name

    def batch_generate(self, *args, **kwargs) -> List[str]:
        """Runs the model to output LLM responses.

        Returns:
            A list of strings.
        """
        raise NotImplementedError(
            "batch_generate is not implemented for this model"
        )

    # Capabilities
    def supports_log_probs(self) -> Union[bool, None]:
        return None

    def supports_temperature(self) -> Union[bool, None]:
        return None

    def supports_multimodal(self) -> Union[bool, None]:
        return None

    def supports_structured_outputs(self) -> Union[bool, None]:
        return None

    def supports_json_mode(self) -> Union[bool, None]:
        return None

    def generate_with_schema(self, *args, schema=None, **kwargs):
        if schema is not None:
            try:
                return self.generate(*args, schema=schema, **kwargs)
            except TypeError:
                pass  # this means provider doesn't accept schema kwarg
        return self.generate(*args, **kwargs)

    async def a_generate_with_schema(self, *args, schema=None, **kwargs):
        if schema is not None:
            try:
                return await self.a_generate(*args, schema=schema, **kwargs)
            except TypeError:
                pass
        return await self.a_generate(*args, **kwargs)


class DeepEvalBaseEmbeddingModel(ABC):
    def __init__(self, model: Optional[str] = None, *args, **kwargs):
        self.name = parse_model_name(model)
        self.model = self.load_model()

    @abstractmethod
    def load_model(self, *args, **kwargs) -> "DeepEvalBaseEmbeddingModel":
        """Loads a model, that will be responsible for generating text embeddings.

        Returns:
            A model object
        """
        pass

    @abstractmethod
    def embed_text(self, *args, **kwargs) -> List[float]:
        """Runs the model to generate text embeddings.

        Returns:
            A list of float.
        """
        pass

    @abstractmethod
    async def a_embed_text(self, *args, **kwargs) -> List[float]:
        """Runs the model to generate text embeddings.

        Returns:
            A list of list of float.
        """
        pass

    @abstractmethod
    def embed_texts(self, *args, **kwargs) -> List[List[float]]:
        """Runs the model to generate list of text embeddings.

        Returns:
            A list of float.
        """
        pass

    @abstractmethod
    async def a_embed_texts(self, *args, **kwargs) -> List[List[float]]:
        """Runs the model to generate list of text embeddings.

        Returns:
            A list of list of float.
        """
        pass

    @abstractmethod
    def get_model_name(self, *args, **kwargs) -> str:
        return self.name



from typing import List, Optional
from pydantic import BaseModel


class ContextualPrecisionVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[ContextualPrecisionVerdict]


class ContextualPrecisionScoreReason(BaseModel):
    reason: str


class InteractionContextualPrecisionScore(BaseModel):
    score: float
    reason: Optional[str]
    verdicts: Optional[List[ContextualPrecisionVerdict]]



import re


class WinograndeTemplate:

    n_shot_examples = [
        {
            "sentence": "Ian volunteered to eat Dennis's menudo after already having a bowl because _ despised eating intestine.",
            "option1": "Ian",
            "option2": "Dennis",
            "answer": "2",
        },
        {
            "sentence": "He never comes to my home, but I always go to his house because the _ is smaller.",
            "option1": "home",
            "option2": "house",
            "answer": "1",
        },
        {
            "sentence": "Kyle doesn't wear leg warmers to bed, while Logan almost always does. _ is more likely to live in a warmer climate.",
            "option1": "Kyle",
            "option2": "Logan",
            "answer": "1",
        },
        {
            "sentence": "The treasury workers took the gold bars off of the trolley and stacked them in the safe until the _ was empty.",
            "option1": "safe",
            "option2": "trolley",
            "answer": "2",
        },
        {
            "sentence": "Emily looked up and saw Patricia racing by overhead, as _ was on the ramp .",
            "option1": "Emily",
            "option2": "Patricia",
            "answer": "2",
        },
    ]

    @staticmethod
    def generate_output(input: str, n_shots: int):
        prompt = ""
        for i in range(n_shots):
            prompt += WinograndeTemplate.format_question(
                WinograndeTemplate.n_shot_examples[i]
            )
        prompt += input
        return prompt

    @staticmethod
    def format_question(data: dict, include_answer: bool = True):
        sentence = data["sentence"]
        option1 = data["option1"]
        option2 = data["option2"]
        answer = data["answer"]
        prompt = f"Sentence: {sentence}\nA. {option1}\nB. {option2}\nAnswer:"
        if include_answer:
            prompt += f"{'A' if answer == '1' else 'B'}\n\n"
        return prompt

    @staticmethod
    def format_answer(data: dict):
        answer = data["answer"]
        return "A" if answer == "1" else "B"



import weakref
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Union,
)
from time import perf_counter
import threading
import functools
import inspect
import asyncio
import random
import atexit
import queue
import uuid
from openai import OpenAI
from rich.console import Console
from rich.progress import Progress

from deepeval.config.settings import get_settings
from deepeval.constants import (
    CONFIDENT_TRACE_VERBOSE,
    CONFIDENT_TRACE_FLUSH,
)
from deepeval.confident.api import Api, Endpoints, HttpMethods, is_confident
from deepeval.metrics import BaseMetric
from deepeval.test_case.llm_test_case import ToolCall
from deepeval.tracing.api import (
    BaseApiSpan,
    SpanApiType,
    TraceApi,
    TraceSpanApiStatus,
)
from deepeval.telemetry import capture_send_trace
from deepeval.tracing.patchers import (
    patch_anthropic_client,
    patch_openai_client,
)
from deepeval.tracing.types import (
    AgentSpan,
    BaseSpan,
    LlmSpan,
    RetrieverSpan,
    SpanType,
    ToolSpan,
    Trace,
    TraceSpanStatus,
    TraceWorkerStatus,
)
from deepeval.tracing.utils import (
    Environment,
    prepare_tool_call_input_parameters,
    replace_self_with_class_name,
    make_json_serializable,
    perf_counter_to_datetime,
    to_zod_compatible_iso,
    tracing_enabled,
    validate_environment,
    validate_sampling_rate,
)
from deepeval.utils import dataclass_to_dict
from deepeval.tracing.context import current_span_context, current_trace_context
from deepeval.tracing.types import TestCaseMetricPair
from deepeval.tracing.api import PromptApi
from deepeval.tracing.trace_test_manager import trace_testing_manager


if TYPE_CHECKING:
    from deepeval.dataset.golden import Golden
    from anthropic import Anthropic

EVAL_DUMMY_SPAN_NAME = "evals_iterator"


class TraceManager:
    def __init__(self):
        self.traces: List[Trace] = []
        self.active_traces: Dict[str, Trace] = {}  # Map of trace_uuid to Trace
        self.active_spans: Dict[str, BaseSpan] = (
            {}
        )  # Map of span_uuid to BaseSpan
        # Map each trace created during evaluation_loop to the Golden that was active
        # when it was started. This lets us evaluate traces against the correct golden
        # since we cannot rely on positional indexing as the order is not guaranteed.
        self.trace_uuid_to_golden: Dict[str, Golden] = {}

        settings = get_settings()
        # Initialize queue and worker thread for trace posting
        self._trace_queue = queue.Queue()
        self._worker_thread = None
        self._min_interval = 0.2  # Minimum time between API calls (seconds)
        self._last_post_time = 0
        self._in_flight_tasks: Set[asyncio.Task[Any]] = set()
        self.task_bindings: "weakref.WeakKeyDictionary[asyncio.Task, dict]" = (
            weakref.WeakKeyDictionary()
        )
        self._flush_enabled = bool(settings.CONFIDENT_TRACE_FLUSH)
        self._daemon = not self._flush_enabled

        # trace manager attributes
        self.confident_api_key = None
        self.custom_mask_fn: Optional[Callable] = None
        self.environment = (
            settings.CONFIDENT_TRACE_ENVIRONMENT
            if settings.CONFIDENT_TRACE_ENVIRONMENT is not None
            else Environment.DEVELOPMENT.value
        )
        validate_environment(self.environment)

        self.sampling_rate = settings.CONFIDENT_TRACE_SAMPLE_RATE
        validate_sampling_rate(self.sampling_rate)
        self.anthropic_client = None
        self.openai_client = None
        self.tracing_enabled = True

        # Evals
        self.evaluating = False
        self.evaluation_loop = False
        self.traces_to_evaluate_order: List[str] = []
        self.traces_to_evaluate: List[Trace] = []
        self.integration_traces_to_evaluate: List[Trace] = []
        self.test_case_metrics: List[TestCaseMetricPair] = []

        # Register an exit handler to warn about unprocessed traces
        atexit.register(self._warn_on_exit)

    def _warn_on_exit(self):
        queue_size = self._trace_queue.qsize()
        in_flight = len(self._in_flight_tasks)
        remaining_tasks = queue_size + in_flight

        if not self._flush_enabled and remaining_tasks > 0:
            self._print_trace_status(
                message=f"WARNING: Exiting with {queue_size + in_flight} abaonded trace(s).",
                trace_worker_status=TraceWorkerStatus.WARNING,
                description=f"Set {CONFIDENT_TRACE_FLUSH}=1 as an environment variable to flush remaining traces to Confident AI.",
            )

    def mask(self, data: Any):
        if self.custom_mask_fn is not None:
            return self.custom_mask_fn(data)
        else:
            return data

    def configure(
        self,
        mask: Optional[Callable] = None,
        environment: Optional[str] = None,
        sampling_rate: Optional[float] = None,
        confident_api_key: Optional[str] = None,
        anthropic_client: Optional["Anthropic"] = None,
        openai_client: Optional[OpenAI] = None,
        tracing_enabled: Optional[bool] = None,
    ) -> None:
        if mask is not None:
            self.custom_mask_fn = mask
        if environment is not None:
            validate_environment(environment)
            self.environment = environment
        if sampling_rate is not None:
            validate_sampling_rate(sampling_rate)
            self.sampling_rate = sampling_rate
        if confident_api_key is not None:
            self.confident_api_key = confident_api_key
        if openai_client is not None:
            self.openai_client = openai_client
            patch_openai_client(openai_client)
        if anthropic_client is not None:
            self.anthropic_client = anthropic_client
            patch_anthropic_client(anthropic_client)
        if tracing_enabled is not None:
            self.tracing_enabled = tracing_enabled

    def start_new_trace(
        self,
        metric_collection: Optional[str] = None,
        trace_uuid: Optional[str] = None,
    ) -> Trace:
        """Start a new trace and set it as the current trace."""
        if trace_uuid is None:
            trace_uuid = str(uuid.uuid4())
        new_trace = Trace(
            uuid=trace_uuid,
            root_spans=[],
            status=TraceSpanStatus.IN_PROGRESS,
            start_time=perf_counter(),
            end_time=None,
            metric_collection=metric_collection,
            confident_api_key=self.confident_api_key,
        )
        self.active_traces[trace_uuid] = new_trace
        self.traces.append(new_trace)
        if self.evaluation_loop:
            self.traces_to_evaluate_order.append(trace_uuid)
            # Associate the current Golden with this trace so we can
            # later evaluate traces against the correct golden, even if more traces
            # are created than goldens or the order interleaves.
            try:
                from deepeval.contextvars import get_current_golden

                current_golden = get_current_golden()
                if current_golden is not None:
                    self.trace_uuid_to_golden[trace_uuid] = current_golden
            except Exception:
                # not much we can do, but if the golden is not there during evaluation
                # we will write out a verbose debug log
                pass
        return new_trace

    def end_trace(self, trace_uuid: str):
        """End a specific trace by its UUID."""

        if trace_uuid in self.active_traces:
            trace = self.active_traces[trace_uuid]
            trace.end_time = (
                perf_counter() if trace.end_time is None else trace.end_time
            )

            # Default to SUCCESS for completed traces
            # This assumes that if a trace completes, it was successful overall
            # Users can manually set the status to ERROR if needed
            if trace.status == TraceSpanStatus.IN_PROGRESS:
                trace.status = TraceSpanStatus.SUCCESS

            if trace_testing_manager.test_name:
                # Trace testing mode is enabled
                # Instead posting the trace to the queue, it will be stored in this global variable
                body = self.create_trace_api(trace).model_dump(
                    by_alias=True, exclude_none=True
                )
                trace_testing_manager.test_dict = make_json_serializable(body)
            #  Post the trace to the server before removing it
            elif not self.evaluating:
                self.post_trace(trace)
            else:
                if self.evaluation_loop:
                    if self.integration_traces_to_evaluate:
                        pass
                    elif self.test_case_metrics:
                        pass
                    elif trace_uuid in self.traces_to_evaluate_order:
                        self.traces_to_evaluate.append(trace)
                        self.traces_to_evaluate.sort(
                            key=lambda t: self.traces_to_evaluate_order.index(
                                t.uuid
                            )
                        )
                else:
                    # print(f"Ending trace: {trace.root_spans}")
                    self.environment = Environment.TESTING
                    if (
                        trace.root_spans
                        and len(trace.root_spans) > 0
                        and trace.root_spans[0].children
                        and len(trace.root_spans[0].children) > 0
                    ):
                        trace.root_spans = [trace.root_spans[0].children[0]]
                    for root_span in trace.root_spans:
                        root_span.parent_uuid = None

            # Remove from active traces
            del self.active_traces[trace_uuid]

    def set_trace_status(self, trace_uuid: str, status: TraceSpanStatus):
        """Manually set the status of a trace."""
        if trace_uuid in self.active_traces:
            trace = self.active_traces[trace_uuid]
            trace.status = status

    def add_span(self, span: BaseSpan):
        """Add a span to the active spans dictionary."""
        self.active_spans[span.uuid] = span

    def remove_span(self, span_uuid: str):
        """Remove a span from the active spans dictionary."""
        if span_uuid in self.active_spans:
            del self.active_spans[span_uuid]

    def add_span_to_trace(self, span: BaseSpan):
        """Add a span to its trace."""
        trace_uuid = span.trace_uuid
        if trace_uuid not in self.active_traces:
            raise ValueError(
                f"Trace with UUID {trace_uuid} does not exist. A span must have a valid trace."
            )

        trace = self.active_traces[trace_uuid]

        # If this is a root span (no parent), add it to the trace's root_spans
        if not span.parent_uuid:
            trace.root_spans.append(span)
        else:
            # This is a child span, find its parent and add it to the parent's children
            parent_span = self.get_span_by_uuid(span.parent_uuid)
            if parent_span:

                if (
                    parent_span.name == EVAL_DUMMY_SPAN_NAME
                ):  # ignored span for evaluation
                    span.parent_uuid = None
                    trace.root_spans.remove(parent_span)
                    trace.root_spans.append(span)
                    return

                parent_span.children.append(span)
            else:
                trace.root_spans.append(span)

    def get_trace_by_uuid(self, trace_uuid: str) -> Optional[Trace]:
        """Get a trace by its UUID."""
        return self.active_traces.get(trace_uuid)

    def get_span_by_uuid(self, span_uuid: str) -> Optional[BaseSpan]:
        """Get a span by its UUID."""
        return self.active_spans.get(span_uuid)

    def get_all_traces(self) -> List[Trace]:
        """Get all traces."""
        return self.traces

    def clear_traces(self):
        """Clear all traces."""
        self.traces = []
        self.active_traces = {}
        self.active_spans = {}

    def get_trace_dict(self, trace: Trace) -> Dict:
        """Convert a trace to a dictionary."""
        return dataclass_to_dict(trace)

    def get_all_traces_dict(self) -> List[Dict]:
        """Get all traces as dictionaries."""
        return [self.get_trace_dict(trace) for trace in self.traces]

    def _print_trace_status(
        self,
        trace_worker_status: TraceWorkerStatus,
        message: str,
        description: Optional[str] = None,
        environment: Optional[str] = None,
    ):
        if get_settings().CONFIDENT_TRACE_VERBOSE and self.evaluating is False:
            console = Console()
            message_prefix = "[dim][Confident AI Trace Log][/dim]"
            if trace_worker_status == TraceWorkerStatus.SUCCESS:
                message = f"[green]{message}[/green]"
            elif trace_worker_status == TraceWorkerStatus.FAILURE:
                message = f"[red]{message}[/red]"
            elif trace_worker_status == TraceWorkerStatus.WARNING:
                message = f"[yellow]{message}[/yellow]"

            env_text = f"[{environment}]" if environment else ""

            if description:
                console.print(
                    message_prefix,
                    env_text,
                    message + ":",
                    description,
                    f"\nTo disable dev logging, set {CONFIDENT_TRACE_VERBOSE}=0 as an environment variable.",
                )
            else:
                console.print(message_prefix, env_text, message)

    def _should_sample_trace(self) -> bool:
        random_number = random.random()
        if random_number > self.sampling_rate:
            rate_str = f"{self.sampling_rate:.2f}"
            self._print_trace_status(
                message=f"Skipped posting trace due to sampling rate ({rate_str})",
                trace_worker_status=TraceWorkerStatus.SUCCESS,
            )
            return False

        return True

    def _ensure_worker_thread_running(self):
        if self._worker_thread is None or not self._worker_thread.is_alive():
            self._worker_thread = threading.Thread(
                target=self._process_trace_queue,
                daemon=self._daemon,
            )
            self._worker_thread.start()

    def post_trace_api(self, trace_api: TraceApi) -> Optional[str]:
        if not tracing_enabled() or not self.tracing_enabled:
            return None

        if not trace_api.confident_api_key:
            if not is_confident() and self.confident_api_key is None:
                self._print_trace_status(
                    message="No Confident AI API key found. Skipping trace posting.",
                    trace_worker_status=TraceWorkerStatus.FAILURE,
                )
                return None

        if not self._should_sample_trace():
            return None

        self._ensure_worker_thread_running()
        self._trace_queue.put(trace_api)

        return

    def post_trace(self, trace: Trace) -> Optional[str]:
        if not tracing_enabled() or not self.tracing_enabled:
            return None

        if not trace.confident_api_key:
            if not is_confident() and self.confident_api_key is None:
                self._print_trace_status(
                    message="No Confident AI API key found. Skipping trace posting.",
                    trace_worker_status=TraceWorkerStatus.FAILURE,
                )
                return None

        if not self._should_sample_trace():
            return None

        # Add the trace to the queue
        self._trace_queue.put(trace)

        # Start the worker thread if it's not already running
        self._ensure_worker_thread_running()

        return

    def _process_trace_queue(self):
        """Worker thread function that processes the trace queue"""
        import threading

        main_thr = threading.main_thread()

        # Create a new event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # buffer for traces that need to be sent after main exits
        remaining_traces: List[TraceApi] = []

        async def _a_send_trace(trace_obj):
            nonlocal remaining_traces
            try:
                # Build API object & payload
                if isinstance(trace_obj, TraceApi):
                    trace_api = trace_obj
                else:
                    trace_api = self.create_trace_api(trace_obj)

                try:
                    body = trace_api.model_dump(
                        by_alias=True,
                        exclude_none=True,
                    )
                except AttributeError:
                    # Pydantic version below 2.0
                    body = trace_api.dict(by_alias=True, exclude_none=True)
                # If the main thread is still alive, send now
                body = make_json_serializable(body)

                if main_thr.is_alive():
                    if trace_api.confident_api_key:
                        api = Api(api_key=trace_api.confident_api_key)
                    else:
                        api = Api(api_key=self.confident_api_key)

                    api_response, link = await api.a_send_request(
                        method=HttpMethods.POST,
                        endpoint=Endpoints.TRACES_ENDPOINT,
                        body=body,
                    )
                    queue_size = self._trace_queue.qsize()
                    in_flight = len(self._in_flight_tasks)
                    status = f"({queue_size} trace{'s' if queue_size!=1 else ''} remaining in queue, {in_flight} in flight)"
                    self._print_trace_status(
                        trace_worker_status=TraceWorkerStatus.SUCCESS,
                        message=f"Successfully posted trace {status}",
                        description=link,
                        environment=self.environment,
                    )
                elif self._flush_enabled:
                    # Main thread gone → to be flushed
                    remaining_traces.append(trace_api)

            except Exception as e:
                queue_size = self._trace_queue.qsize()
                in_flight = len(self._in_flight_tasks)
                status = f"({queue_size} trace{'s' if queue_size!=1 else ''} remaining in queue, {in_flight} in flight)"
                self._print_trace_status(
                    trace_worker_status=TraceWorkerStatus.FAILURE,
                    message=f"Error posting trace {status}",
                    description=str(e),
                )
            finally:
                task = asyncio.current_task()
                if task:
                    self._in_flight_tasks.discard(task)

        async def async_worker():
            # Continue while user code is running or work remains
            while (
                main_thr.is_alive()
                or not self._trace_queue.empty()
                or self._in_flight_tasks
            ):
                try:
                    trace = self._trace_queue.get(block=True, timeout=1.0)

                    # rate-limit
                    now = perf_counter()
                    elapsed = now - self._last_post_time
                    if elapsed < self._min_interval:
                        await asyncio.sleep(self._min_interval - elapsed)
                    self._last_post_time = perf_counter()

                    # schedule async send
                    task = asyncio.create_task(_a_send_trace(trace))
                    self._in_flight_tasks.add(task)
                    self._trace_queue.task_done()

                except queue.Empty:
                    await asyncio.sleep(0.1)
                    continue
                except Exception as e:
                    self._print_trace_status(
                        message="Error in worker",
                        trace_worker_status=TraceWorkerStatus.FAILURE,
                        description=str(e),
                    )
                    await asyncio.sleep(1.0)

        try:
            loop.run_until_complete(async_worker())
        finally:
            # Drain any pending tasks
            pending = asyncio.all_tasks(loop=loop)
            if pending:
                loop.run_until_complete(
                    asyncio.gather(*pending, return_exceptions=True)
                )
            self.flush_traces(remaining_traces)
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    def flush_traces(self, remaining_traces: List[TraceApi]):
        if not tracing_enabled() or not self.tracing_enabled:
            return

        self._print_trace_status(
            TraceWorkerStatus.WARNING,
            message=f"Flushing {len(remaining_traces)} remaining trace(s)",
        )
        for trace_api in remaining_traces:
            with capture_send_trace():
                try:
                    try:
                        body = trace_api.model_dump(
                            by_alias=True,
                            exclude_none=True,
                        )
                    except AttributeError:
                        # Pydantic version below 2.0
                        body = trace_api.dict(by_alias=True, exclude_none=True)

                    body = make_json_serializable(body)
                    if trace_api.confident_api_key:
                        api = Api(api_key=trace_api.confident_api_key)
                    else:
                        api = Api(api_key=self.confident_api_key)

                    _, link = api.send_request(
                        method=HttpMethods.POST,
                        endpoint=Endpoints.TRACES_ENDPOINT,
                        body=body,
                    )
                    qs = self._trace_queue.qsize()
                    self._print_trace_status(
                        trace_worker_status=TraceWorkerStatus.SUCCESS,
                        message=f"Successfully posted trace ({qs} traces remaining in queue, 1 in flight)",
                        description=link,
                        environment=self.environment,
                    )
                except Exception as e:
                    qs = self._trace_queue.qsize()
                    self._print_trace_status(
                        trace_worker_status=TraceWorkerStatus.FAILURE,
                        message="Error flushing remaining trace(s)",
                        description=str(e),
                    )

    def create_nested_spans_dict(self, span: BaseSpan) -> Dict[str, Any]:
        api_span = self._convert_span_to_api_span(span)
        trace_dict = api_span.__dict__.copy()

        # Remove specific keys
        for key in (
            "uuid",
            "trace_uuid",
            "parent_uuid",
            "end_time",
            "start_time",
            "status",
            "llm_test_case",
            "metrics_data",
            "metric_collection",
            "metadata",
        ):
            trace_dict.pop(key, None)

        # Remove all keys with None values
        trace_dict = {k: v for k, v in trace_dict.items() if v is not None}

        trace_dict["children"] = []
        for child in span.children or []:
            child_api_span = self.create_nested_spans_dict(child)
            trace_dict["children"].append(child_api_span)

        return trace_dict

    def create_trace_api(self, trace: Trace) -> TraceApi:
        # Initialize empty lists for each span type
        base_spans = []
        agent_spans = []
        llm_spans = []
        retriever_spans = []
        tool_spans = []

        # Process all spans in the trace iteratively
        span_stack = list(trace.root_spans)  # Start with root spans

        while span_stack:
            span = span_stack.pop()

            # Convert BaseSpan to BaseApiSpan
            api_span = self._convert_span_to_api_span(span)

            # Categorize spans by type
            if isinstance(span, AgentSpan):
                agent_spans.append(api_span)
            elif isinstance(span, LlmSpan):
                llm_spans.append(api_span)
            elif isinstance(span, RetrieverSpan):
                retriever_spans.append(api_span)
            elif isinstance(span, ToolSpan):
                tool_spans.append(api_span)
            else:
                base_spans.append(api_span)

            # Add children to the stack for processing
            if span.children:
                span_stack.extend(span.children)

        # Convert perf_counter values to ISO 8601 strings
        start_time = (
            to_zod_compatible_iso(perf_counter_to_datetime(trace.start_time))
            if trace.start_time
            else None
        )
        end_time = (
            to_zod_compatible_iso(perf_counter_to_datetime(trace.end_time))
            if trace.end_time
            else None
        )

        return TraceApi(
            uuid=trace.uuid,
            baseSpans=base_spans,
            agentSpans=agent_spans,
            llmSpans=llm_spans,
            retrieverSpans=retriever_spans,
            toolSpans=tool_spans,
            startTime=start_time,
            endTime=end_time,
            metadata=trace.metadata,
            name=trace.name,
            tags=trace.tags,
            threadId=trace.thread_id,
            userId=trace.user_id,
            input=trace.input,
            output=trace.output,
            metricCollection=trace.metric_collection,
            retrievalContext=trace.retrieval_context,
            context=trace.context,
            expectedOutput=trace.expected_output,
            toolsCalled=trace.tools_called,
            expectedTools=trace.expected_tools,
            confident_api_key=trace.confident_api_key,
            environment=(
                self.environment if not trace.environment else trace.environment
            ),
            status=(
                TraceSpanApiStatus.SUCCESS
                if trace.status == TraceSpanStatus.SUCCESS
                else TraceSpanApiStatus.ERRORED
            ),
        )

    def _convert_span_to_api_span(self, span: BaseSpan) -> BaseApiSpan:
        # Determine span type
        if isinstance(span, AgentSpan):
            span_type = SpanApiType.AGENT
        elif isinstance(span, LlmSpan):
            span_type = SpanApiType.LLM
        elif isinstance(span, RetrieverSpan):
            span_type = SpanApiType.RETRIEVER
        elif isinstance(span, ToolSpan):
            span_type = SpanApiType.TOOL
        else:
            span_type = SpanApiType.BASE

        # Initialize input and output fields
        input_data = span.input
        output_data = span.output

        # Convert perf_counter values to ISO 8601 strings
        start_time = (
            to_zod_compatible_iso(perf_counter_to_datetime(span.start_time))
            if span.start_time
            else None
        )
        end_time = (
            to_zod_compatible_iso(perf_counter_to_datetime(span.end_time))
            if span.end_time
            else None
        )

        from deepeval.evaluate.utils import create_metric_data

        # Create the base API span
        api_span = BaseApiSpan(
            uuid=span.uuid,
            name=span.name,
            status=span.status.value,
            type=span_type,
            parentUuid=span.parent_uuid,
            startTime=start_time,
            endTime=end_time,
            input=input_data,
            output=output_data,
            metadata=span.metadata,
            error=span.error,
            metricCollection=span.metric_collection,
            metricsData=(
                [create_metric_data(metric) for metric in span.metrics]
                if span.metrics
                else None
            ),
            retrievalContext=span.retrieval_context,
            context=span.context,
            expectedOutput=span.expected_output,
            toolsCalled=span.tools_called,
            expectedTools=span.expected_tools,
        )

        # Add type-specific attributes
        if isinstance(span, AgentSpan):
            api_span.available_tools = span.available_tools
            api_span.agent_handoffs = span.agent_handoffs
        elif isinstance(span, ToolSpan):
            api_span.description = span.description
        elif isinstance(span, RetrieverSpan):
            api_span.embedder = span.embedder
            api_span.top_k = span.top_k
            api_span.chunk_size = span.chunk_size
        elif isinstance(span, LlmSpan):
            api_span.model = span.model
            alias = span.prompt.alias if span.prompt else None
            version = span.prompt.version if span.prompt else None
            api_span.prompt = PromptApi(alias=alias, version=version)
            api_span.cost_per_input_token = span.cost_per_input_token
            api_span.cost_per_output_token = span.cost_per_output_token
            api_span.input_token_count = span.input_token_count
            api_span.output_token_count = span.output_token_count

            processed_token_intervals = {}
            if span.token_intervals:
                for key, value in span.token_intervals.items():
                    time = to_zod_compatible_iso(
                        perf_counter_to_datetime(key),
                        microsecond_precision=True,
                    )
                    processed_token_intervals[time] = value
                api_span.token_intervals = processed_token_intervals

        return api_span


trace_manager = TraceManager()

########################################################
### Observer #############################################
########################################################


class Observer:
    def __init__(
        self,
        span_type: Union[
            Literal["agent", "llm", "retriever", "tool"], str, None
        ],
        func_name: str,
        metrics: Optional[Union[List[str], List[BaseMetric]]] = None,
        metric_collection: Optional[str] = None,
        _progress: Optional[Progress] = None,
        _pbar_callback_id: Optional[int] = None,
        **kwargs,
    ):
        self.start_time: float
        self.end_time: float
        self.status: TraceSpanStatus
        self.error: Optional[str] = None
        self.uuid: str = str(uuid.uuid4())
        # Initialize trace_uuid and parent_uuid as None, they will be set in __enter__
        self.trace_uuid: Optional[str] = None
        self.parent_uuid: Optional[str] = None

        # Separate observe kwargs and function kwargs
        self.observe_kwargs = kwargs.get("observe_kwargs", {})
        self.function_kwargs = kwargs.get("function_kwargs", {})
        self.result = None

        self.name: str = self.observe_kwargs.get("name", func_name)
        self.prompt = self.observe_kwargs.get("prompt", None)
        self.metrics = metrics
        self.metric_collection = metric_collection
        self.span_type: Optional[SpanType] = span_type
        self._progress = _progress
        self._pbar_callback_id = _pbar_callback_id
        self.update_span_properties: Optional[Callable] = None

    def __enter__(self):
        """Enter the tracer context, creating a new span and setting up parent-child relationships."""
        self.start_time = perf_counter()

        # Get the current span from the context
        parent_span = current_span_context.get()

        # Determine trace_uuid and parent_uuid before creating the span instance
        if parent_span:
            self.parent_uuid = parent_span.uuid
            self.trace_uuid = parent_span.trace_uuid
        else:
            current_trace = current_trace_context.get()
            if current_trace:
                self.trace_uuid = current_trace.uuid
            else:
                trace = trace_manager.start_new_trace(
                    metric_collection=self.metric_collection
                )
                self.trace_uuid = trace.uuid
                current_trace_context.set(trace)

        # Now create the span instance with the correct trace_uuid and parent_uuid
        span_instance = self.create_span_instance()

        # stash call arguments so they are available during the span lifetime
        setattr(span_instance, "_function_kwargs", self.function_kwargs)

        # Add the span to active spans and to its trace
        trace_manager.add_span(span_instance)
        trace_manager.add_span_to_trace(span_instance)

        # Set this span as the current span in the context
        current_span_context.set(span_instance)
        if (
            parent_span
            and parent_span.progress is not None
            and parent_span.pbar_callback_id is not None
        ):
            self._progress = parent_span.progress
            self._pbar_callback_id = parent_span.pbar_callback_id

        try:
            import asyncio

            task = asyncio.current_task()
        except Exception:
            task = None

        if task is not None:
            binding = trace_manager.task_bindings.get(task) or {}
            # record the trace the task is working on
            binding["trace_uuid"] = span_instance.trace_uuid
            # only set root_span_uuid when this span is a root. Don't do this for child or we will override our record.
            if (
                span_instance.parent_uuid is None
                and "root_span_uuid" not in binding
            ):
                binding["root_span_uuid"] = span_instance.uuid
            trace_manager.task_bindings[task] = binding

        if self._progress is not None and self._pbar_callback_id is not None:
            span_instance.progress = self._progress
            span_instance.pbar_callback_id = self._pbar_callback_id

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the tracer context, updating the span status and handling trace completion."""

        end_time = perf_counter()
        # Get the current span from the context instead of looking it up by UUID
        current_span = current_span_context.get()

        if not current_span or current_span.uuid != self.uuid:
            print(
                f"Error: Current span in context does not match the span being exited. Expected UUID: {self.uuid}, Got: {current_span.uuid if current_span else 'None'}"
            )
            return

        current_span.end_time = end_time
        if exc_type is not None:
            current_span.status = TraceSpanStatus.ERRORED
            current_span.error = str(exc_val)
        else:
            current_span.status = TraceSpanStatus.SUCCESS

        if self.update_span_properties is not None:
            self.update_span_properties(current_span)

        if current_span.input is None:
            current_span.input = trace_manager.mask(self.function_kwargs)
        if current_span.output is None:
            current_span.output = trace_manager.mask(self.result)

        if (
            isinstance(current_span, LlmSpan)
            and self.prompt
            and not current_span.prompt
        ):
            current_span.prompt = self.prompt

        if not current_span.tools_called:
            # check any tool span children
            for child in current_span.children:
                if isinstance(child, ToolSpan):
                    current_span.tools_called = current_span.tools_called or []
                    current_span.tools_called.append(
                        ToolCall(
                            name=child.name,
                            description=child.description,
                            input_parameters=prepare_tool_call_input_parameters(
                                child.input
                            ),
                            output=child.output,
                        )
                    )

        trace_manager.remove_span(self.uuid)
        if current_span.parent_uuid:
            parent_span = trace_manager.get_span_by_uuid(
                current_span.parent_uuid
            )
            if parent_span:
                current_span_context.set(parent_span)
            else:
                current_span_context.set(None)
        else:
            current_trace = current_trace_context.get()
            if current_trace.input is None:
                current_trace.input = self.function_kwargs
            if current_trace.output is None:
                current_trace.output = self.result
            if current_span.status == TraceSpanStatus.ERRORED:
                current_trace.status = TraceSpanStatus.ERRORED
            if current_trace and current_trace.uuid == current_span.trace_uuid:
                other_active_spans = [
                    span
                    for span in trace_manager.active_spans.values()
                    if span.trace_uuid == current_span.trace_uuid
                ]

                if not other_active_spans:
                    trace_manager.end_trace(current_span.trace_uuid)
                    current_trace_context.set(None)

            current_span_context.set(None)

        if self._progress is not None and self._pbar_callback_id is not None:
            self._progress.update(self._pbar_callback_id, advance=1)

    def create_span_instance(self):
        """Create a span instance based on the span type."""

        span_kwargs = {
            "uuid": self.uuid,
            "trace_uuid": self.trace_uuid,
            "parent_uuid": self.parent_uuid,
            "start_time": self.start_time,
            "end_time": None,
            "status": TraceSpanStatus.SUCCESS,
            "children": [],
            "name": self.name,
            # "metadata": None,
            "input": None,
            "output": None,
            "metrics": self.metrics,
            "metric_collection": self.metric_collection,
        }

        if self.span_type == SpanType.AGENT.value:
            available_tools = self.observe_kwargs.get("available_tools", [])
            agent_handoffs = self.observe_kwargs.get("agent_handoffs", [])

            return AgentSpan(
                **span_kwargs,
                available_tools=available_tools,
                agent_handoffs=agent_handoffs,
            )
        elif self.span_type == SpanType.LLM.value:
            model = self.observe_kwargs.get("model", None)
            cost_per_input_token = self.observe_kwargs.get(
                "cost_per_input_token", None
            )
            cost_per_output_token = self.observe_kwargs.get(
                "cost_per_output_token", None
            )
            return LlmSpan(
                **span_kwargs,
                model=model,
                cost_per_input_token=cost_per_input_token,
                cost_per_output_token=cost_per_output_token,
            )
        elif self.span_type == SpanType.RETRIEVER.value:
            embedder = self.observe_kwargs.get("embedder", None)
            return RetrieverSpan(**span_kwargs, embedder=embedder)

        elif self.span_type == SpanType.TOOL.value:
            return ToolSpan(**span_kwargs, **self.observe_kwargs)
        else:
            return BaseSpan(**span_kwargs)


########################################################
### Decorator ##########################################
########################################################


def observe(
    _func: Optional[Callable] = None,
    *,
    metrics: Optional[List[BaseMetric]] = None,
    metric_collection: Optional[str] = None,
    type: Optional[
        Union[Literal["agent", "llm", "retriever", "tool"], str]
    ] = None,
    **observe_kwargs,
):
    """
    Decorator to trace a function as a span.

    Args:
        span_type: The type of span to create (AGENT, LLM, RETRIEVER, TOOL, or custom string)
        **observe_kwargs: Additional arguments to pass to the Observer

    Returns:
        A decorator function that wraps the original function with a Observer
    """

    def decorator(func):
        func_name = func.__name__  # Get func_name outside wrappers

        # Async generator function
        if inspect.isasyncgenfunction(func):

            @functools.wraps(func)
            def asyncgen_wrapper(*args, **func_kwargs):

                sig = inspect.signature(func)
                bound = sig.bind(*args, **func_kwargs)
                bound.apply_defaults()

                complete_kwargs = dict(bound.arguments)
                if "self" in complete_kwargs:
                    complete_kwargs["self"] = replace_self_with_class_name(
                        complete_kwargs["self"]
                    )
                observer_kwargs = {
                    "observe_kwargs": observe_kwargs,
                    "function_kwargs": complete_kwargs,
                }

                observer = Observer(
                    type,
                    metrics=metrics,
                    metric_collection=metric_collection,
                    func_name=func_name,
                    **observer_kwargs,
                )
                observer.__enter__()
                agen = func(*args, **func_kwargs)

                async def gen():
                    try:
                        async for chunk in agen:
                            yield chunk
                        observer.__exit__(None, None, None)
                    except Exception as e:
                        observer.__exit__(type(e), e, e.__traceback__)
                        raise

                return gen()

            setattr(asyncgen_wrapper, "_is_deepeval_observed", True)
            return asyncgen_wrapper

        # Sync generator function
        if inspect.isgeneratorfunction(func):

            @functools.wraps(func)
            def gen_wrapper(*args, **func_kwargs):

                sig = inspect.signature(func)
                bound = sig.bind(*args, **func_kwargs)
                bound.apply_defaults()
                complete_kwargs = dict(bound.arguments)

                if "self" in complete_kwargs:
                    complete_kwargs["self"] = replace_self_with_class_name(
                        complete_kwargs["self"]
                    )
                observer_kwargs = {
                    "observe_kwargs": observe_kwargs,
                    "function_kwargs": make_json_serializable(complete_kwargs),
                }

                observer = Observer(
                    type,
                    metrics=metrics,
                    metric_collection=metric_collection,
                    func_name=func_name,
                    **observer_kwargs,
                )
                observer.__enter__()
                original_gen = func(*args, **func_kwargs)

                def gen():
                    try:
                        yield from original_gen
                        observer.__exit__(None, None, None)
                    except Exception as e:
                        observer.__exit__(type(e), e, e.__traceback__)
                        raise

                return gen()

            setattr(gen_wrapper, "_is_deepeval_observed", True)
            return gen_wrapper

        if asyncio.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **func_kwargs):
                # func_name = func.__name__ # Removed from here
                sig = inspect.signature(func)
                bound_args = sig.bind(*args, **func_kwargs)
                bound_args.apply_defaults()

                # Construct complete kwargs dictionary & pass all kwargs with consistent naming
                complete_kwargs = dict(bound_args.arguments)
                observer_kwargs = {
                    "observe_kwargs": observe_kwargs,
                    "function_kwargs": complete_kwargs,  # Now contains all args mapped to their names
                }
                with Observer(
                    type,
                    metrics=metrics,
                    metric_collection=metric_collection,
                    func_name=func_name,
                    **observer_kwargs,
                ) as observer:
                    # Call the original function
                    result = await func(*args, **func_kwargs)
                    # Capture the result
                    observer.result = result
                    return result

            # Set the marker attribute on the wrapper
            setattr(async_wrapper, "_is_deepeval_observed", True)
            return async_wrapper
        else:

            @functools.wraps(func)
            def wrapper(*args, **func_kwargs):
                # func_name = func.__name__ # Removed from here
                sig = inspect.signature(func)
                bound_args = sig.bind(*args, **func_kwargs)
                bound_args.apply_defaults()
                complete_kwargs = dict(bound_args.arguments)

                if "self" in complete_kwargs:
                    complete_kwargs["self"] = replace_self_with_class_name(
                        complete_kwargs["self"]
                    )

                observer_kwargs = {
                    "observe_kwargs": observe_kwargs,
                    "function_kwargs": make_json_serializable(
                        complete_kwargs
                    ),  # serilaizing it before it goes to trace api and raises circular reference error
                }
                with Observer(
                    type,
                    metrics=metrics,
                    metric_collection=metric_collection,
                    func_name=func_name,
                    **observer_kwargs,
                ) as observer:
                    # Call the original function
                    result = func(*args, **func_kwargs)
                    # Capture the result
                    observer.result = make_json_serializable(
                        result
                    )  # serilaizing it before it goes to trace api and raises circular reference error
                    return result

            # Set the marker attribute on the wrapper
            setattr(wrapper, "_is_deepeval_observed", True)
            return wrapper

    if _func is not None and callable(_func):
        return decorator(_func)

    return decorator



# Internal GEPA constants - not exposed to users
GEPA_MIN_DELTA: float = 0.0
GEPA_TIE_TOLERANCE: float = 1e-9
GEPA_REWRITE_INSTRUCTION_MAX_CHARS: int = 4096

# Internal MIPROV2 constants - not exposed to users
MIPROV2_MIN_DELTA: float = 0.0
MIPROV2_REWRITE_INSTRUCTION_MAX_CHARS: int = 4096
MIPROV2_DEFAULT_NUM_CANDIDATES: int = 10
MIPROV2_DEFAULT_NUM_TRIALS: int = 20
MIPROV2_DEFAULT_MINIBATCH_SIZE: int = 25
MIPROV2_DEFAULT_MINIBATCH_FULL_EVAL_STEPS: int = 10
MIPROV2_DEFAULT_MAX_BOOTSTRAPPED_DEMOS: int = 4
MIPROV2_DEFAULT_MAX_LABELED_DEMOS: int = 4
MIPROV2_DEFAULT_NUM_DEMO_SETS: int = 5

# Internal SIMBA constants - not exposed to users
SIMBA_DEMO_INPUT_MAX_CHARS: int = 256



from deepeval.openai_agents.callback_handler import DeepEvalTracingProcessor
from deepeval.openai_agents.agent import DeepEvalAgent as Agent
from deepeval.openai_agents.patch import function_tool

# from deepeval.openai_agents.runner import Runner

__all__ = ["DeepEvalTracingProcessor", "Agent", "function_tool"]



from .handler import instrument_crewai
from .subs import (
    DeepEvalCrew as Crew,
    DeepEvalAgent as Agent,
    DeepEvalLLM as LLM,
)
from .tool import tool

__all__ = ["instrument_crewai", "Crew", "Agent", "LLM", "tool"]



class ARCTemplate:
    n_shot_examples = [
        {
            "id": "Mercury_7220990",
            "question": "Which factor will most likely cause a person to develop a fever?",
            "choices": {
                "text": [
                    "a leg muscle relaxing after exercise",
                    "a bacterial population in the bloodstream",
                    "several viral particles on the skin",
                    "carbohydrates being digested in the stomach",
                ],
                "label": ["A", "B", "C", "D"],
            },
            "answerKey": "B",
        },
        {
            "id": "MCAS_2007_8_5189",
            "question": "Lichens are symbiotic organisms made of green algae and fungi. What do the green algae supply to the fungi in this symbiotic relationship?",
            "choices": {
                "text": ["carbon dioxide", "food", "protection", "water"],
                "label": ["A", "B", "C", "D"],
            },
            "answerKey": "B",
        },
        {
            "id": "Mercury_SC_401169",
            "question": "When a switch is used in an electrical circuit, the switch can",
            "choices": {
                "text": [
                    "cause the charge to build.",
                    "increase and decrease the voltage.",
                    "cause the current to change direction.",
                    "stop and start the flow of current.",
                ],
                "label": ["A", "B", "C", "D"],
            },
            "answerKey": "D",
        },
        {
            "id": "MCAS_2004_8_27",
            "question": "Which of the following is an example of an assistive device?",
            "choices": {
                "text": [
                    "contact lens",
                    "motorcycle",
                    "raincoat",
                    "coffee pot",
                ],
                "label": ["A", "B", "C", "D"],
            },
            "answerKey": "A",
        },
        {
            "id": "NYSEDREGENTS_2006_8_10",
            "question": "Rocks are classified as igneous, metamorphic, or sedimentary according to",
            "choices": {
                "text": [
                    "their color",
                    "their shape",
                    "how they formed",
                    "the minerals they contain",
                ],
                "label": ["1", "2", "3", "4"],
            },
            "answerKey": "3",
        },
    ]

    @staticmethod
    def generate_output(input: str, n_shots: int):
        prompt = ""
        for i in range(n_shots):
            prompt += ARCTemplate.format_question(
                ARCTemplate.n_shot_examples[i]
            )
        prompt += input
        return prompt

    @staticmethod
    def format_question(data: dict, include_answer: bool = True):
        prompt = data["question"]
        texts = data["choices"]["text"]
        labels = data["choices"]["label"]
        for i in range(len(labels)):
            prompt += "\n{}. {}".format(labels[i], texts[i])
        prompt += "\nAnswer: "
        if include_answer:
            prompt += " {}\n\n".format(data["answerKey"])
        return prompt

    @staticmethod
    def format_answer(data: dict):
        return data["answerKey"]



from deepeval.test_case import LLMTestCase
import textwrap


class MCPUseMetricTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
    """

    @staticmethod
    def get_mcp_argument_correctness_prompt(
        test_case: LLMTestCase,
        available_primitives: str,
        primitives_used: str,
    ):
        return textwrap.dedent(
            f"""Evaluate whether the arguments passed to each tool (primitive) used by the agent were appropriate and correct for the intended purpose. Focus on whether the input types, formats, and contents match the expectations of the tools and are suitable given the user's request.

            {MCPUseMetricTemplate.multimodal_rules}

            You must return a JSON object with exactly two fields: 'score' and 'reason'.

            Scoring:
            - 'score' is a float between 0 and 1 inclusive.
            - Use intermediate values (e.g., 0.25, 0.5, 0.75) to reflect partial correctness, such as when argument types were correct but content was misaligned with intent.
            - 'reason' should clearly explain whether the arguments passed to tools were well-formed, appropriate, and aligned with the tool’s expected inputs and the user’s request.

            IMPORTANT:
            - Assume the selected tools themselves were appropriate (do NOT judge tool selection).
            - Focus ONLY on:
            - Whether the correct arguments were passed to each tool (e.g., types, structure, semantics).
            - Whether any required arguments were missing or malformed.
            - Whether extraneous, irrelevant, or incorrect values were included.
            - Refer to 'available_primitives' to understand expected argument formats and semantics.

            CHAIN OF THOUGHT:
            1. Understand the user’s request from 'test_case.input'.
            2. Review the arguments passed to each tool in 'primitives_used' (structure, content, type).
            3. Compare the arguments with what each tool in 'available_primitives' expects.
            4. Determine whether each tool was used with suitable and valid inputs, including values aligned with the task.
            5. Do NOT evaluate tool choice or output quality — only input correctness for the tools used.

            You must return only a valid JSON object. Do not include any explanation or text outside the JSON.

            -----------------
            User Input:
            {test_case.input}

            Agent Visible Output:
            {test_case.actual_output}

            Available Primitives (with expected arguments and signatures):
            {available_primitives}

            Primitives Used by Agent (with arguments passed):
            {primitives_used}

            Example Output:
            {{
                "score": 0.5,
                "reason": "The agent passed arguments of the correct type to all tools, but one tool received an input that did not match the user's intent and another had a missing required field."
            }}

            JSON:
            """
        )

    @staticmethod
    def get_primitive_correctness_prompt(
        test_case: LLMTestCase,
        available_primitives: str,
        primitives_used: str,
    ):
        return textwrap.dedent(
            f"""Evaluate whether the tools (primitives) selected and used by the agent were appropriate and correct for fulfilling the user’s request. Base your judgment on the user input, the agent’s visible output, and the tools that were available to the agent. You must return a JSON object with exactly two fields: 'score' and 'reason'.

            {MCPUseMetricTemplate.multimodal_rules}

            Scoring:
            - 'score' is a float between 0 and 1 inclusive.
            - Use intermediate values (e.g., 0.25, 0.5, 0.75) to reflect cases where the tools used were partially correct, suboptimal, or only somewhat relevant.
            - 'reason' should clearly explain how appropriate and correct the chosen primitives were, considering both the user's request and the output.

            IMPORTANT:
            - Focus only on tool selection and usage — not the quality of the final output.
            - Assume that 'available_primitives' contains the only tools the agent could have used.
            - Consider whether the agent:
            - Chose the correct tool(s) for the task.
            - Avoided unnecessary or incorrect tool calls.
            - Missed a more appropriate tool when one was available.
            - Multiple valid tool combinations may exist — give credit when one reasonable strategy is used effectively.

            CHAIN OF THOUGHT:
            1. Determine what the user was asking for from 'test_case.input'.
            2. Evaluate whether the tools in 'primitives_used' were appropriate for achieving that goal.
            3. Consider the list of 'available_primitives' to judge if better options were missed or if poor tools were unnecessarily used.
            4. Ignore whether the tool *worked* — focus only on whether it was the *right tool to use*.

            You must return only a valid JSON object. Do not include any explanation or text outside the JSON.

            -----------------
            User Input:
            {test_case.input}

            Agent Visible Output:
            {test_case.actual_output}

            Available Tools:
            {available_primitives}

            Tools Used by Agent:
            {primitives_used}

            Example Output:
            {{
                "score": 0.75,
                "reason": "The agent used a relevant tool to address the user's request, but a more specific tool was available and would have been more efficient."
            }}

            JSON:
            """
        )



from pydantic import BaseModel


class UserInputAndTools(BaseModel):
    user_messages: str
    assistant_messages: str
    tools_called: str
    available_tools: str
    tools_used: bool


class ToolSelectionScore(BaseModel):
    score: float
    reason: str


class ArgumentCorrectnessScore(BaseModel):
    score: float
    reason: str



from deepeval.models.base_model import DeepEvalBaseLLM
from abc import ABC, abstractmethod
from typing import List, TypeVar, Generic, List, Optional
from pydantic import BaseModel

from deepeval.dataset import Golden


class DeepEvalBaseBenchmarkResult(BaseModel):
    overall_accuracy: float


T = TypeVar("T")


class DeepEvalBaseBenchmark(ABC, Generic[T]):
    def __init__(self, dataset: Optional["Dataset"] = None):
        from datasets import Dataset

        self.tasks: List[T] = []
        self.dataset = dataset

    @abstractmethod
    def load_benchmark_dataset(self, *args, **kwargs) -> List[Golden]:
        """Load the benchmark dataset and initialize tasks."""
        raise NotImplementedError

    @abstractmethod
    def evaluate(
        self, model: DeepEvalBaseLLM, *args, **kwargs
    ) -> DeepEvalBaseBenchmarkResult:
        raise NotImplementedError



from deepeval.confident.api import Api, Endpoints, HttpMethods
from deepeval.tracing.context import current_trace_context
from deepeval.tracing.offline_evals.api import EvaluateThreadRequestBody


def evaluate_thread(
    thread_id: str, metric_collection: str, overwrite_metrics: bool = False
):
    trace = current_trace_context.get()
    api_key = None
    if trace:
        api_key = trace.confident_api_key
    api = Api(api_key=api_key)

    evaluate_thread_request_body = EvaluateThreadRequestBody(
        metricCollection=metric_collection,
        overwriteMetrics=overwrite_metrics,
    )
    try:
        body = evaluate_thread_request_body.model_dump(
            by_alias=True,
            exclude_none=True,
        )
    except AttributeError:
        # Pydantic version below 2.0
        body = evaluate_thread_request_body.dict(
            by_alias=True, exclude_none=True
        )

    api.send_request(
        method=HttpMethods.POST,
        endpoint=Endpoints.EVALUATE_THREAD_ENDPOINT,
        body=body,
        url_params={"threadId": thread_id},
    )


async def a_evaluate_thread(
    thread_id: str, metric_collection: str, overwrite_metrics: bool = False
):
    trace = current_trace_context.get()
    api_key = None
    if trace:
        api_key = trace.confident_api_key
    api = Api(api_key=api_key)

    evaluate_thread_request_body = EvaluateThreadRequestBody(
        metricCollection=metric_collection,
        overwriteMetrics=overwrite_metrics,
    )
    try:
        body = evaluate_thread_request_body.model_dump(
            by_alias=True,
            exclude_none=True,
        )
    except AttributeError:
        # Pydantic version below 2.0
        body = evaluate_thread_request_body.dict(
            by_alias=True, exclude_none=True
        )

    await api.a_send_request(
        method=HttpMethods.POST,
        endpoint=Endpoints.EVALUATE_THREAD_ENDPOINT,
        body=body,
        url_params={"threadId": thread_id},
    )



from deepeval.contextvars import get_current_golden
from .dataset import EvaluationDataset
from .golden import Golden, ConversationalGolden


__all__ = [
    "EvaluationDataset",
    "Golden",
    "ConversationalGolden",
    "get_current_golden",
]



from __future__ import annotations

import inspect
from typing import Any, Callable, Optional, List
from deepeval.tracing.context import current_span_context
from deepeval.tracing.types import AgentSpan, ToolSpan
from deepeval.tracing.utils import make_json_serializable
from deepeval.tracing import observe
from deepeval.tracing.tracing import Observer
from deepeval.metrics import BaseMetric
from deepeval.prompt import Prompt
from deepeval.tracing.types import LlmSpan
from functools import wraps

try:
    from agents import function_tool as _agents_function_tool  # type: ignore
    from deepeval.openai_agents.extractors import parse_response_output
    from agents.run import AgentRunner
    from agents.run import SingleStepResult
    from agents.models.interface import Model
    from agents import Agent
except Exception:
    pass


def _wrap_with_observe(
    func: Callable[..., Any],
    metrics: Optional[str] = None,
    metric_collection: Optional[str] = None,
) -> Callable[..., Any]:
    if getattr(func, "_is_deepeval_observed", False):
        return func

    if inspect.iscoroutinefunction(func):

        @wraps(func)
        async def observed(*args: Any, **kwargs: Any) -> Any:
            current_span = current_span_context.get()
            if isinstance(current_span, ToolSpan):
                current_span.metrics = metrics
                current_span.metric_collection = metric_collection
            return await func(*args, **kwargs)

    else:

        @wraps(func)
        def observed(*args: Any, **kwargs: Any) -> Any:
            current_span = current_span_context.get()
            if isinstance(current_span, ToolSpan):
                current_span.metrics = metrics
                current_span.metric_collection = metric_collection
            return func(*args, **kwargs)

    setattr(observed, "_is_deepeval_observed", True)
    try:
        observed.__signature__ = inspect.signature(func)  # type: ignore[attr-defined]
    except Exception:
        pass
    return observed


def function_tool(
    func: Optional[Callable[..., Any]] = None, /, *args: Any, **kwargs: Any
) -> Any:
    metrics = kwargs.pop("metrics", None)
    metric_collection = kwargs.pop("metric_collection", None)

    if _agents_function_tool is None:
        raise RuntimeError(
            "agents.function_tool is not available. Please install agents via your package manager"
        )

    if callable(func):

        wrapped = _wrap_with_observe(
            func,
            metrics=metrics,
            metric_collection=metric_collection,
        )
        return _agents_function_tool(wrapped, *args, **kwargs)

    def decorator(real_func: Callable[..., Any]) -> Any:

        wrapped = _wrap_with_observe(
            real_func,
            metrics=metrics,
            metric_collection=metric_collection,
        )
        return _agents_function_tool(wrapped, *args, **kwargs)

    return decorator


_PATCHED_DEFAULT_RUN_SINGLE_TURN = False
_PATCHED_DEFAULT_RUN_SINGLE_TURN_STREAMED = False
_PATCHED_DEFAULT_GET_MODEL = False


class _ObservedModel(Model):
    def __init__(
        self,
        inner: Model,
        llm_metric_collection: str = None,
        llm_metrics: List[BaseMetric] = None,
        confident_prompt: Prompt = None,
    ) -> None:
        self._inner = inner
        self._llm_metric_collection = llm_metric_collection
        self._llm_metrics = llm_metrics
        self._confident_prompt = confident_prompt

    def __getattr__(self, name: str) -> Any:
        return getattr(self._inner, name)

    async def get_response(
        self,
        *args,
        **kwargs,
    ):
        with Observer(
            span_type="llm",
            func_name="LLM",
            observe_kwargs={"model": "temp_model"},
            metrics=self._llm_metrics,
            metric_collection=self._llm_metric_collection,
        ):
            result = await self._inner.get_response(
                *args,
                **kwargs,
            )
            llm_span: LlmSpan = current_span_context.get()
            llm_span.prompt = self._confident_prompt

        return result

    def stream_response(
        self,
        *args,
        **kwargs,
    ):

        async def _gen():
            observer = Observer(
                span_type="llm",
                func_name="LLM",
                observe_kwargs={"model": "temp_model"},
                metrics=self._llm_metrics,
                metric_collection=self._llm_metric_collection,
            )
            observer.__enter__()

            llm_span: LlmSpan = current_span_context.get()
            llm_span.prompt = self._confident_prompt

            try:
                async for event in self._inner.stream_response(
                    *args,
                    **kwargs,
                ):
                    yield event
            except Exception as e:
                observer.__exit__(type(e), e, e.__traceback__)
                raise
            finally:
                observer.__exit__(None, None, None)

        return _gen()


def patch_default_agent_run_single_turn():
    global _PATCHED_DEFAULT_RUN_SINGLE_TURN
    if _PATCHED_DEFAULT_RUN_SINGLE_TURN:
        return

    original_run_single_turn = AgentRunner._run_single_turn

    @classmethod
    async def patched_run_single_turn(cls, *args, **kwargs):
        res: SingleStepResult = await original_run_single_turn.__func__(
            cls, *args, **kwargs
        )
        try:
            if isinstance(res, SingleStepResult):
                agent_span = current_span_context.get()
                if isinstance(agent_span, AgentSpan):
                    _set_agent_metrics(kwargs.get("agent", None), agent_span)
                    if agent_span.input is None:
                        _pre_step_items_raw_list = [
                            item.raw_item for item in res.pre_step_items
                        ]
                        agent_span.input = (
                            make_json_serializable(_pre_step_items_raw_list)
                            if _pre_step_items_raw_list
                            else make_json_serializable(res.original_input)
                        )
                    agent_span.output = parse_response_output(
                        res.model_response.output
                    )
        except Exception:
            pass
        return res

    AgentRunner._run_single_turn = patched_run_single_turn
    _PATCHED_DEFAULT_RUN_SINGLE_TURN = True  # type: ignore


def patch_default_agent_run_single_turn_streamed():
    global _PATCHED_DEFAULT_RUN_SINGLE_TURN_STREAMED
    if _PATCHED_DEFAULT_RUN_SINGLE_TURN_STREAMED:
        return

    original_run_single_turn_streamed = AgentRunner._run_single_turn_streamed

    @classmethod
    async def patched_run_single_turn_streamed(cls, *args, **kwargs):

        res: SingleStepResult = (
            await original_run_single_turn_streamed.__func__(
                cls, *args, **kwargs
            )
        )
        try:
            if isinstance(res, SingleStepResult):
                agent_span = current_span_context.get()
                if isinstance(agent_span, AgentSpan):
                    _set_agent_metrics(
                        kwargs.get("agent", None), agent_span
                    )  # TODO: getting no agent
                    if agent_span.input is None:
                        _pre_step_items_raw_list = [
                            item.raw_item for item in res.pre_step_items
                        ]
                        agent_span.input = (
                            make_json_serializable(_pre_step_items_raw_list)
                            if _pre_step_items_raw_list
                            else make_json_serializable(res.original_input)
                        )
                    agent_span.output = parse_response_output(
                        res.model_response.output
                    )
        except Exception:
            pass
        return res

    AgentRunner._run_single_turn_streamed = patched_run_single_turn_streamed
    _PATCHED_DEFAULT_RUN_SINGLE_TURN_STREAMED = True  # type: ignore


def patch_default_agent_runner_get_model():
    global _PATCHED_DEFAULT_GET_MODEL
    if _PATCHED_DEFAULT_GET_MODEL:
        return

    original_get_model_cm = AgentRunner._get_model
    try:
        original_get_model = original_get_model_cm.__func__
    except AttributeError:
        original_get_model = (
            original_get_model_cm  # fallback (non-classmethod edge case)
        )

    def patched_get_model(cls, *args, **kwargs) -> Model:
        model = original_get_model(cls, *args, **kwargs)

        agent = (
            kwargs.get("agent")
            if "agent" in kwargs
            else (args[0] if args else None)
        )
        if agent is None:
            return model

        if isinstance(model, _ObservedModel):
            return model

        llm_metrics = getattr(agent, "llm_metrics", None)
        llm_metric_collection = getattr(agent, "llm_metric_collection", None)
        confident_prompt = getattr(agent, "confident_prompt", None)
        return _ObservedModel(
            inner=model,
            llm_metric_collection=llm_metric_collection,
            llm_metrics=llm_metrics,
            confident_prompt=confident_prompt,
        )

    # Preserve basic metadata and mark as patched
    patched_get_model.__name__ = original_get_model.__name__
    patched_get_model.__doc__ = original_get_model.__doc__

    AgentRunner._get_model = classmethod(patched_get_model)
    _PATCHED_DEFAULT_GET_MODEL = True


def _set_agent_metrics(agent: Agent, agent_span: AgentSpan) -> None:
    try:
        if agent is None or agent_span is None:
            return
        agent_metrics = getattr(agent, "agent_metrics", None)
        agent_metric_collection = getattr(
            agent, "agent_metric_collection", None
        )

        if agent_metrics is not None:
            agent_span.metrics = agent_metrics
        if agent_metric_collection is not None:
            agent_span.metric_collection = agent_metric_collection
    except Exception:
        # Be conservative: never break the run on metrics propagation
        pass



import aiohttp
import requests
from typing import Any, Dict, List, Optional, Union
from pydantic import AnyUrl, SecretStr

from deepeval.errors import DeepEvalError
from deepeval.config.settings import get_settings
from deepeval.models.utils import (
    require_secret_api_key,
)
from deepeval.test_case import MLLMImage
from deepeval.utils import check_if_multimodal, convert_to_multi_modal_array
from deepeval.models import DeepEvalBaseLLM
from deepeval.utils import require_param


def _request_timeout_seconds() -> float:
    timeout = float(get_settings().DEEPEVAL_PER_ATTEMPT_TIMEOUT_SECONDS or 0)
    return timeout if timeout > 0 else 30.0


class PortkeyModel(DeepEvalBaseLLM):
    def __init__(
        self,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        base_url: Optional[AnyUrl] = None,
        provider: Optional[str] = None,
        generation_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        settings = get_settings()
        model = model or settings.PORTKEY_MODEL_NAME

        if api_key is not None:
            # keep it secret, keep it safe from serializings, logging and alike
            self.api_key: Optional[SecretStr] = SecretStr(api_key)
        else:
            self.api_key = settings.PORTKEY_API_KEY

        if base_url is not None:
            base_url = str(base_url).rstrip("/")
        elif settings.PORTKEY_BASE_URL is not None:
            base_url = str(settings.PORTKEY_BASE_URL).rstrip("/")

        provider = provider or settings.PORTKEY_PROVIDER_NAME

        # validation
        model = require_param(
            model,
            provider_label="Portkey",
            env_var_name="PORTKEY_MODEL_NAME",
            param_hint="model",
        )

        self.base_url = require_param(
            base_url,
            provider_label="Portkey",
            env_var_name="PORTKEY_BASE_URL",
            param_hint="base_url",
        )

        self.provider = require_param(
            provider,
            provider_label="Portkey",
            env_var_name="PORTKEY_PROVIDER_NAME",
            param_hint="provider",
        )
        # Keep sanitized kwargs for client call to strip legacy keys
        self.kwargs = kwargs
        self.generation_kwargs = generation_kwargs or {}
        super().__init__(model)

    def _headers(self) -> Dict[str, str]:
        api_key = require_secret_api_key(
            self.api_key,
            provider_label="Portkey",
            env_var_name="PORTKEY_API_KEY",
            param_hint="`api_key` to PortkeyModel(...)",
        )

        headers = {
            "Content-Type": "application/json",
            "x-portkey-api-key": api_key,
        }
        if self.provider:
            headers["x-portkey-provider"] = self.provider
        return headers

    def _payload(self, prompt: str) -> Dict[str, Any]:
        if check_if_multimodal(prompt):
            prompt = convert_to_multi_modal_array(input=prompt)
            content = self.generate_content(prompt)
        else:
            content = [{"type": "text", "text": prompt}]
        payload = {
            "model": self.name,
            "messages": [{"role": "user", "content": content}],
        }
        if self.generation_kwargs:
            payload.update(self.generation_kwargs)
        return payload

    def generate_content(
        self, multimodal_input: List[Union[str, MLLMImage]] = []
    ):
        content = []
        for element in multimodal_input:
            if isinstance(element, str):
                content.append({"type": "text", "text": element})
            elif isinstance(element, MLLMImage):
                if element.url and not element.local:
                    content.append(
                        {
                            "type": "image_url",
                            "image_url": {"url": element.url},
                        }
                    )
                else:
                    element.ensure_images_loaded()
                    data_uri = (
                        f"data:{element.mimeType};base64,{element.dataBase64}"
                    )
                    content.append(
                        {
                            "type": "image_url",
                            "image_url": {"url": data_uri},
                        }
                    )
        return content

    def _extract_content(self, data: Dict[str, Any]) -> str:
        choices: Union[List[Dict[str, Any]], None] = data.get("choices")
        if not choices:
            raise DeepEvalError("Portkey response did not include any choices.")
        message = choices[0].get("message", {})
        content: Union[str, List[Dict[str, Any]], None] = message.get("content")
        if isinstance(content, str):
            return content
        if isinstance(content, list):
            return "".join(part.get("text", "") for part in content)
        return ""

    def generate(self, prompt: str) -> str:

        try:
            response = requests.post(
                f"{self.base_url}/chat/completions",
                json=self._payload(prompt),
                headers=self._headers(),
                timeout=_request_timeout_seconds(),
            )
            response.raise_for_status()
        except requests.HTTPError as error:
            body: Union[str, Dict[str, Any]]
            try:
                body = response.json()
            except Exception:
                body = response.text
            raise DeepEvalError(
                f"Portkey request failed with status {response.status_code}: {body}"
            ) from error
        except requests.RequestException as error:
            raise DeepEvalError(f"Portkey request failed: {error}") from error
        return self._extract_content(response.json())

    async def a_generate(self, prompt: str) -> str:

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.base_url}/chat/completions",
                json=self._payload(prompt),
                headers=self._headers(),
                timeout=_request_timeout_seconds(),
            ) as response:
                if response.status >= 400:
                    body = await response.text()
                    raise DeepEvalError(
                        f"Portkey request failed with status {response.status}: {body}"
                    )
                data = await response.json()
                return self._extract_content(data)

    def load_model(self):
        return None

    def get_model_name(self):
        return f"{self.name} (Portkey)"

    def supports_multimodal(self):
        return True



# MIPROv2 - Multiprompt Instruction PRoposal Optimizer Version 2
#
# This implementation follows the original MIPROv2 paper and DSPy implementation:
# https://arxiv.org/pdf/2406.11695
# https://dspy.ai/api/optimizers/MIPROv2/
#
# The algorithm works in two phases:
#
#   1. PROPOSAL PHASE:
#      a) Generate N diverse instruction candidates upfront
#      b) Bootstrap few-shot demonstration sets from training data
#
#   2. OPTIMIZATION PHASE: Use Bayesian Optimization (Optuna TPE) to search
#      over the joint space of (instruction_candidate, demo_set). Each trial:
#      - Samples an instruction candidate index
#      - Samples a demo set index
#      - Renders the prompt with demos
#      - Evaluates on a minibatch of examples
#      - Uses the score to guide the Bayesian surrogate model
#
# Periodic full evaluation is performed every `minibatch_full_eval_steps`
# to get accurate scores on the complete validation set.


from __future__ import annotations
import asyncio
import uuid
import random
import time
import logging
from typing import (
    Dict,
    List,
    Tuple,
    TYPE_CHECKING,
    Union,
    Optional,
    Callable,
)

try:
    import optuna
    from optuna.samplers import TPESampler

    OPTUNA_AVAILABLE = True
except ImportError:
    OPTUNA_AVAILABLE = False
    optuna = None
    TPESampler = None

from deepeval.models.base_model import DeepEvalBaseLLM
from deepeval.errors import DeepEvalError
from deepeval.optimizer.utils import Aggregator, mean_of_all
from deepeval.optimizer.types import (
    PromptConfiguration,
    PromptConfigurationId,
    ModuleId,
    ScoreTable,
    OptimizationReport,
    RunnerStatusType,
    RunnerStatusCallback,
)
from deepeval.optimizer.scorer.base import BaseScorer
from deepeval.optimizer.algorithms.base import BaseAlgorithm
from deepeval.optimizer.utils import build_prompt_config_snapshots
from deepeval.prompt.prompt import Prompt
from deepeval.optimizer.algorithms.miprov2.proposer import InstructionProposer
from deepeval.optimizer.algorithms.miprov2.bootstrapper import (
    DemoBootstrapper,
    DemoSet,
    render_prompt_with_demos,
)
from deepeval.optimizer.algorithms.configs import (
    MIPROV2_DEFAULT_NUM_CANDIDATES,
    MIPROV2_DEFAULT_NUM_TRIALS,
    MIPROV2_DEFAULT_MINIBATCH_SIZE,
    MIPROV2_DEFAULT_MINIBATCH_FULL_EVAL_STEPS,
    MIPROV2_DEFAULT_MAX_BOOTSTRAPPED_DEMOS,
    MIPROV2_DEFAULT_MAX_LABELED_DEMOS,
    MIPROV2_DEFAULT_NUM_DEMO_SETS,
)

if TYPE_CHECKING:
    from deepeval.dataset.golden import Golden, ConversationalGolden


# Suppress Optuna's verbose logging
logging.getLogger("optuna").setLevel(logging.WARNING)


class MIPROV2(BaseAlgorithm):
    """
    MIPROv2 (Multiprompt Instruction PRoposal Optimizer Version 2)

    A prompt optimizer that uses Bayesian Optimization to find the best
    combination of instruction and few-shot demonstrations. Follows the
    original MIPROv2 paper approach.

    The optimization process:
    1. Generate N diverse instruction candidates upfront
    2. Bootstrap M demo sets from training examples
    3. Use Optuna's TPE sampler for Bayesian Optimization over (instruction, demos)
    4. Each trial evaluates a combination on a minibatch
    5. Periodically evaluate the best combination on the full dataset

    Parameters
    ----------
    num_candidates : int
        Number of instruction candidates to propose. Default is 10.
    num_trials : int
        Number of Bayesian Optimization trials. Default is 20.
    minibatch_size : int
        Number of examples per minibatch evaluation. Default is 25.
    minibatch_full_eval_steps : int
        Evaluate best on full dataset every N trials. Default is 10.
    max_bootstrapped_demos : int
        Maximum bootstrapped demos per demo set. Default is 4.
    max_labeled_demos : int
        Maximum labeled demos (from expected_output) per set. Default is 4.
    num_demo_sets : int
        Number of demo sets to create. Default is 5.
    random_seed : int, optional
        RNG seed for reproducibility. If None, derived from time.time_ns().
    aggregate_instances : Aggregator
        Function to aggregate per-instance scores. Default is mean_of_all.
    scorer : BaseScorer, optional
        Scorer for evaluating prompts. Set by PromptOptimizer.
    """

    name = "MIPROv2"
    SINGLE_MODULE_ID: ModuleId = "__module__"

    def __init__(
        self,
        num_candidates: int = MIPROV2_DEFAULT_NUM_CANDIDATES,
        num_trials: int = MIPROV2_DEFAULT_NUM_TRIALS,
        minibatch_size: int = MIPROV2_DEFAULT_MINIBATCH_SIZE,
        minibatch_full_eval_steps: int = MIPROV2_DEFAULT_MINIBATCH_FULL_EVAL_STEPS,
        max_bootstrapped_demos: int = MIPROV2_DEFAULT_MAX_BOOTSTRAPPED_DEMOS,
        max_labeled_demos: int = MIPROV2_DEFAULT_MAX_LABELED_DEMOS,
        num_demo_sets: int = MIPROV2_DEFAULT_NUM_DEMO_SETS,
        random_seed: Optional[int] = None,
        aggregate_instances: Aggregator = mean_of_all,
        scorer: Optional[BaseScorer] = None,
    ) -> None:
        if not OPTUNA_AVAILABLE:
            raise DeepEvalError(
                "MIPROv2 requires the 'optuna' package for Bayesian Optimization. "
                "Install it with: pip install optuna"
            )

        # Validate parameters
        if num_candidates < 1:
            raise ValueError("num_candidates must be >= 1")
        if num_trials < 1:
            raise ValueError("num_trials must be >= 1")
        if minibatch_size < 1:
            raise ValueError("minibatch_size must be >= 1")
        if minibatch_full_eval_steps < 1:
            raise ValueError("minibatch_full_eval_steps must be >= 1")
        if max_bootstrapped_demos < 0:
            raise ValueError("max_bootstrapped_demos must be >= 0")
        if max_labeled_demos < 0:
            raise ValueError("max_labeled_demos must be >= 0")
        if num_demo_sets < 1:
            raise ValueError("num_demo_sets must be >= 1")

        self.num_candidates = num_candidates
        self.num_trials = num_trials
        self.minibatch_size = minibatch_size
        self.minibatch_full_eval_steps = minibatch_full_eval_steps
        self.max_bootstrapped_demos = max_bootstrapped_demos
        self.max_labeled_demos = max_labeled_demos
        self.num_demo_sets = num_demo_sets
        self.aggregate_instances = aggregate_instances
        self.scorer = scorer

        # Random seed handling
        if random_seed is None:
            random_seed = time.time_ns() % (2**31)
        self.random_seed = random_seed
        self.random_state = random.Random(random_seed)

        # Runtime state
        self.reset_state()

        # Callbacks and models (set by PromptOptimizer)
        self.status_callback: Optional[RunnerStatusCallback] = None
        self.optimizer_model: Optional["DeepEvalBaseLLM"] = None

        # Lazy-loaded components
        self._proposer: Optional[InstructionProposer] = None
        self._bootstrapper: Optional[DemoBootstrapper] = None

    ##############
    # Public API #
    ##############

    def execute(
        self,
        prompt: Prompt,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> Tuple[Prompt, OptimizationReport]:
        """
        Synchronous MIPROv2 optimization.

        Phase 1: Propose instruction candidates + Bootstrap demo sets
        Phase 2: Use Bayesian Optimization to find the best combination
        """
        self._validate_inputs(goldens)
        self._ensure_scorer()
        self._ensure_proposer()
        self._ensure_bootstrapper()
        self.reset_state()

        # Phase 1a: Propose instruction candidates
        self._update_status("Phase 1: Proposing instruction candidates...", 0)
        instruction_candidates = self._proposer.propose(
            prompt=prompt,
            goldens=goldens,
            num_candidates=self.num_candidates,
        )
        self._register_instruction_candidates(instruction_candidates)

        # Phase 1b: Bootstrap demo sets
        self._update_status(
            "Phase 1: Bootstrapping few-shot demonstrations...", 0
        )
        self._demo_sets = self._bootstrapper.bootstrap(
            prompt=prompt,
            goldens=goldens,
            generate_fn=self._create_generate_fn(),
        )
        self._update_status(f"Bootstrapped {len(self._demo_sets)} demo sets", 0)

        # Phase 2: Bayesian Optimization over (instruction, demos)
        self._update_status("Phase 2: Starting Bayesian Optimization...", 0)
        best_instr_idx, best_demo_idx = self._run_bayesian_optimization(goldens)

        # Final full evaluation if not already done
        config_key = (best_instr_idx, best_demo_idx)
        if config_key not in self._full_eval_cache:
            best_config = self._get_config_by_index(best_instr_idx)
            best_demo_set = self._demo_sets[best_demo_idx]
            self._full_evaluate(best_config, best_demo_set, goldens)

        # Build report
        best = self._best_by_aggregate()
        return self._build_result(best)

    async def a_execute(
        self,
        prompt: Prompt,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> Tuple[Prompt, OptimizationReport]:
        """
        Asynchronous MIPROv2 optimization.
        """
        self._validate_inputs(goldens)
        self._ensure_scorer()
        self._ensure_proposer()
        self._ensure_bootstrapper()
        self.reset_state()

        # Phase 1: Run proposal and bootstrapping concurrently
        self._update_status(
            "Phase 1: Proposing candidates & bootstrapping demos...", 0
        )

        instruction_candidates, demo_sets = await asyncio.gather(
            self._proposer.a_propose(
                prompt=prompt,
                goldens=goldens,
                num_candidates=self.num_candidates,
            ),
            self._bootstrapper.a_bootstrap(
                prompt=prompt,
                goldens=goldens,
                a_generate_fn=self._create_async_generate_fn(),
            ),
        )

        self._register_instruction_candidates(instruction_candidates)
        self._demo_sets = demo_sets
        self._update_status(
            f"Generated {len(instruction_candidates)} candidates, {len(self._demo_sets)} demo sets",
            0,
        )

        # Phase 2: Bayesian Optimization
        self._update_status("Phase 2: Starting Bayesian Optimization...", 0)
        best_instr_idx, best_demo_idx = await self._a_run_bayesian_optimization(
            goldens
        )

        # Final full evaluation if not already done
        config_key = (best_instr_idx, best_demo_idx)
        if config_key not in self._full_eval_cache:
            best_config = self._get_config_by_index(best_instr_idx)
            best_demo_set = self._demo_sets[best_demo_idx]
            await self._a_full_evaluate(best_config, best_demo_set, goldens)

        # Build report
        best = self._best_by_aggregate()
        return self._build_result(best)

    ###################
    # State & Helpers #
    ###################

    def reset_state(self) -> None:
        """Reset optimization state for a new run."""
        self.optimization_id = str(uuid.uuid4())
        self.prompt_configurations_by_id: Dict[
            PromptConfigurationId, PromptConfiguration
        ] = {}
        self.parents_by_id: Dict[
            PromptConfigurationId, Optional[PromptConfigurationId]
        ] = {}
        self.pareto_score_table: ScoreTable = {}

        # Candidate tracking
        self._instruction_candidates: List[PromptConfiguration] = []
        self._demo_sets: List[DemoSet] = []

        # Score tracking: (instr_idx, demo_idx) -> list of minibatch scores
        self._combination_scores: Dict[Tuple[int, int], List[float]] = {}

        # Full eval cache: (instr_idx, demo_idx) -> config_id
        self._full_eval_cache: Dict[Tuple[int, int], PromptConfigurationId] = {}

        # Trial tracking
        self._trial_history: List[Dict] = []
        self._best_trial_key: Tuple[int, int] = (0, 0)
        self._best_trial_score: float = float("-inf")

    def _validate_inputs(
        self,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> None:
        """Validate input parameters."""
        if len(goldens) < 1:
            raise DeepEvalError(
                "MIPROv2 prompt optimization requires at least 1 golden, but "
                f"received {len(goldens)}. Provide at least one golden to run "
                "the optimizer."
            )

    def _ensure_scorer(self) -> None:
        """Ensure scorer is configured."""
        if self.scorer is None:
            raise DeepEvalError(
                "MIPROv2 requires a `scorer`. "
                "Construct one in PromptOptimizer and assign it to `runner.scorer`."
            )

    def _ensure_proposer(self) -> None:
        """Lazily initialize the instruction proposer."""
        if self._proposer is None:
            if self.optimizer_model is None:
                raise DeepEvalError(
                    "MIPROv2 requires an `optimizer_model` for instruction proposal. "
                    "Set it via PromptOptimizer."
                )
            self._proposer = InstructionProposer(
                optimizer_model=self.optimizer_model,
                random_state=self.random_state,
            )

    def _ensure_bootstrapper(self) -> None:
        """Lazily initialize the demo bootstrapper."""
        if self._bootstrapper is None:
            self._bootstrapper = DemoBootstrapper(
                max_bootstrapped_demos=self.max_bootstrapped_demos,
                max_labeled_demos=self.max_labeled_demos,
                num_demo_sets=self.num_demo_sets,
                random_state=self.random_state,
            )

    def _create_generate_fn(
        self,
    ) -> Callable[[Prompt, Union["Golden", "ConversationalGolden"]], str]:
        """Create a sync generate function for bootstrapping."""

        def generate_fn(
            prompt: Prompt,
            golden: Union["Golden", "ConversationalGolden"],
        ) -> str:
            # Create a temporary config for generation
            temp_config = PromptConfiguration.new(
                prompts={self.SINGLE_MODULE_ID: prompt}
            )
            return self.scorer.generate(temp_config.prompts, golden)

        return generate_fn

    def _create_async_generate_fn(self) -> Callable:
        """Create an async generate function for bootstrapping."""

        async def a_generate_fn(
            prompt: Prompt,
            golden: Union["Golden", "ConversationalGolden"],
        ) -> str:
            temp_config = PromptConfiguration.new(
                prompts={self.SINGLE_MODULE_ID: prompt}
            )
            return await self.scorer.a_generate(temp_config.prompts, golden)

        return a_generate_fn

    def _register_instruction_candidates(
        self, candidates: List[Prompt]
    ) -> None:
        """Register all instruction candidates as configurations."""
        for i, prompt in enumerate(candidates):
            config = PromptConfiguration.new(
                prompts={self.SINGLE_MODULE_ID: prompt},
                parent=None if i == 0 else self._instruction_candidates[0].id,
            )
            self._instruction_candidates.append(config)
            self.prompt_configurations_by_id[config.id] = config
            self.parents_by_id[config.id] = config.parent

    def _get_config_by_index(self, idx: int) -> PromptConfiguration:
        """Get configuration by instruction candidate index."""
        return self._instruction_candidates[idx]

    def _draw_minibatch(
        self,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> Union[List["Golden"], List["ConversationalGolden"]]:
        """Sample a minibatch from goldens."""
        n = len(goldens)
        if n <= 0:
            return []
        size = min(self.minibatch_size, n)
        return [goldens[self.random_state.randrange(0, n)] for _ in range(size)]

    def _render_config_with_demos(
        self,
        config: PromptConfiguration,
        demo_set: DemoSet,
    ) -> PromptConfiguration:
        """Create a new config with demos rendered into the prompt."""
        base_prompt = config.prompts[self.SINGLE_MODULE_ID]
        rendered_prompt = render_prompt_with_demos(
            prompt=base_prompt,
            demo_set=demo_set,
            max_demos=self.max_bootstrapped_demos + self.max_labeled_demos,
        )

        # Create a new config with the rendered prompt
        rendered_config = PromptConfiguration.new(
            prompts={self.SINGLE_MODULE_ID: rendered_prompt},
            parent=config.id,
        )
        return rendered_config

    ############################
    # Bayesian Optimization    #
    ############################

    def _run_bayesian_optimization(
        self,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> Tuple[int, int]:
        """
        Run Bayesian Optimization using Optuna's TPE sampler.
        Returns the (instruction_idx, demo_set_idx) of the best combination.
        """
        num_instructions = len(self._instruction_candidates)
        num_demo_sets = len(self._demo_sets)

        # Create Optuna study with TPE sampler
        sampler = TPESampler(seed=self.random_seed)
        study = optuna.create_study(
            direction="maximize",
            sampler=sampler,
        )

        def objective(trial: "optuna.Trial") -> float:
            # Sample instruction and demo set indices
            instr_idx = trial.suggest_int("instr_idx", 0, num_instructions - 1)
            demo_idx = trial.suggest_int("demo_idx", 0, num_demo_sets - 1)

            # Get the configuration and demo set
            config = self._get_config_by_index(instr_idx)
            demo_set = self._demo_sets[demo_idx]

            # Render prompt with demos
            rendered_config = self._render_config_with_demos(config, demo_set)

            # Draw minibatch and score
            minibatch = self._draw_minibatch(goldens)
            score = self.scorer.score_minibatch(rendered_config, minibatch)

            # Track scores for this combination
            combo_key = (instr_idx, demo_idx)
            if combo_key not in self._combination_scores:
                self._combination_scores[combo_key] = []
            self._combination_scores[combo_key].append(score)

            # Update best tracking
            if score > self._best_trial_score:
                self._best_trial_score = score
                self._best_trial_key = combo_key

            # Record trial
            trial_num = len(self._trial_history) + 1
            self._trial_history.append(
                {
                    "trial": trial_num,
                    "instr_idx": instr_idx,
                    "demo_idx": demo_idx,
                    "score": score,
                }
            )

            # Progress update
            demo_info = (
                f"{len(demo_set.demos)} demos" if demo_set.demos else "0-shot"
            )
            self._update_status(
                f"Trial {trial_num}/{self.num_trials} - "
                f"Instr {instr_idx}, {demo_info} - Score: {score:.4f}",
                trial_num,
            )

            # Periodic full evaluation
            if trial_num % self.minibatch_full_eval_steps == 0:
                best_instr, best_demo = self._best_trial_key
                if (best_instr, best_demo) not in self._full_eval_cache:
                    best_config = self._get_config_by_index(best_instr)
                    best_demo_set = self._demo_sets[best_demo]
                    self._full_evaluate(best_config, best_demo_set, goldens)

            return score

        # Run optimization
        study.optimize(
            objective,
            n_trials=self.num_trials,
            show_progress_bar=False,
        )

        # Return the best combination
        return (
            study.best_params["instr_idx"],
            study.best_params["demo_idx"],
        )

    async def _a_run_bayesian_optimization(
        self,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> Tuple[int, int]:
        """
        Async version of Bayesian Optimization.
        """
        num_instructions = len(self._instruction_candidates)
        num_demo_sets = len(self._demo_sets)

        sampler = TPESampler(seed=self.random_seed)
        study = optuna.create_study(
            direction="maximize",
            sampler=sampler,
        )

        for trial_num in range(1, self.num_trials + 1):
            trial = study.ask()

            # Sample indices
            instr_idx = trial.suggest_int("instr_idx", 0, num_instructions - 1)
            demo_idx = trial.suggest_int("demo_idx", 0, num_demo_sets - 1)

            # Get config and demos
            config = self._get_config_by_index(instr_idx)
            demo_set = self._demo_sets[demo_idx]
            rendered_config = self._render_config_with_demos(config, demo_set)

            # Score on minibatch
            minibatch = self._draw_minibatch(goldens)
            score = await self.scorer.a_score_minibatch(
                rendered_config, minibatch
            )

            # Track scores
            combo_key = (instr_idx, demo_idx)
            if combo_key not in self._combination_scores:
                self._combination_scores[combo_key] = []
            self._combination_scores[combo_key].append(score)

            # Update best
            if score > self._best_trial_score:
                self._best_trial_score = score
                self._best_trial_key = combo_key

            # Record trial
            self._trial_history.append(
                {
                    "trial": trial_num,
                    "instr_idx": instr_idx,
                    "demo_idx": demo_idx,
                    "score": score,
                }
            )

            # Tell Optuna the result
            study.tell(trial, score)

            # Progress update
            demo_info = (
                f"{len(demo_set.demos)} demos" if demo_set.demos else "0-shot"
            )
            self._update_status(
                f"Trial {trial_num}/{self.num_trials} - "
                f"Instr {instr_idx}, {demo_info} - Score: {score:.4f}",
                trial_num,
            )

            # Periodic full evaluation
            if trial_num % self.minibatch_full_eval_steps == 0:
                best_instr, best_demo = self._best_trial_key
                if (best_instr, best_demo) not in self._full_eval_cache:
                    best_config = self._get_config_by_index(best_instr)
                    best_demo_set = self._demo_sets[best_demo]
                    await self._a_full_evaluate(
                        best_config, best_demo_set, goldens
                    )

        return (
            study.best_params["instr_idx"],
            study.best_params["demo_idx"],
        )

    ############################
    # Full Evaluation          #
    ############################

    def _full_evaluate(
        self,
        config: PromptConfiguration,
        demo_set: DemoSet,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> None:
        """Perform full evaluation on all goldens."""
        # Find the indices for this combination
        instr_idx = self._instruction_candidates.index(config)
        demo_idx = self._demo_sets.index(demo_set)
        combo_key = (instr_idx, demo_idx)

        if combo_key in self._full_eval_cache:
            return

        # Render with demos
        rendered_config = self._render_config_with_demos(config, demo_set)

        # Register the rendered config
        self.prompt_configurations_by_id[rendered_config.id] = rendered_config
        self.parents_by_id[rendered_config.id] = config.id

        # Score on full set
        scores = self.scorer.score_pareto(rendered_config, goldens)
        self.pareto_score_table[rendered_config.id] = scores

        # Cache the result
        self._full_eval_cache[combo_key] = rendered_config.id

    async def _a_full_evaluate(
        self,
        config: PromptConfiguration,
        demo_set: DemoSet,
        goldens: Union[List["Golden"], List["ConversationalGolden"]],
    ) -> None:
        """Async full evaluation."""
        instr_idx = self._instruction_candidates.index(config)
        demo_idx = self._demo_sets.index(demo_set)
        combo_key = (instr_idx, demo_idx)

        if combo_key in self._full_eval_cache:
            return

        rendered_config = self._render_config_with_demos(config, demo_set)
        self.prompt_configurations_by_id[rendered_config.id] = rendered_config
        self.parents_by_id[rendered_config.id] = config.id

        scores = await self.scorer.a_score_pareto(rendered_config, goldens)
        self.pareto_score_table[rendered_config.id] = scores
        self._full_eval_cache[combo_key] = rendered_config.id

    ############################
    # Result Building          #
    ############################

    def _best_by_aggregate(self) -> PromptConfiguration:
        """Return the best candidate based on full evaluation scores."""
        if not self.pareto_score_table:
            # Fall back to best by trial scores
            best_instr, best_demo = self._best_trial_key
            config = self._get_config_by_index(best_instr)
            demo_set = self._demo_sets[best_demo]
            return self._render_config_with_demos(config, demo_set)

        best_id: Optional[PromptConfigurationId] = None
        best_score = float("-inf")

        for config_id, scores in self.pareto_score_table.items():
            agg_score = self.aggregate_instances(scores)
            if agg_score > best_score:
                best_score = agg_score
                best_id = config_id

        if best_id is None:
            best_instr, best_demo = self._best_trial_key
            config = self._get_config_by_index(best_instr)
            demo_set = self._demo_sets[best_demo]
            return self._render_config_with_demos(config, demo_set)

        return self.prompt_configurations_by_id[best_id]

    def _build_result(
        self,
        best: PromptConfiguration,
    ) -> Tuple[Prompt, OptimizationReport]:
        """Build the optimization result."""
        prompt_config_snapshots = build_prompt_config_snapshots(
            self.prompt_configurations_by_id
        )

        report = OptimizationReport(
            optimization_id=self.optimization_id,
            best_id=best.id,
            accepted_iterations=self._trial_history,
            pareto_scores=self.pareto_score_table,
            parents=self.parents_by_id,
            prompt_configurations=prompt_config_snapshots,
        )

        return best.prompts[self.SINGLE_MODULE_ID], report

    ############################
    # Status Updates           #
    ############################

    def _update_status(self, message: str, step: int) -> None:
        """Send status update via callback."""
        if self.status_callback is not None:
            self.status_callback(
                RunnerStatusType.PROGRESS,
                step_index=step,
                total_steps=self.num_trials,
                detail=message,
            )



from typing import Optional
from dataclasses import dataclass

from deepeval.test_run.test_run import TestRunResultDisplay


@dataclass
class AsyncConfig:
    run_async: bool = True
    throttle_value: float = 0
    max_concurrent: int = 20

    def __post_init__(self):
        if self.max_concurrent < 1:
            raise ValueError("'max_concurrent' must be at least 1")
        if self.throttle_value < 0:
            raise ValueError("'throttle_value' must be at least 0")


@dataclass
class DisplayConfig:
    show_indicator: bool = True
    print_results: bool = True
    verbose_mode: Optional[bool] = None
    display_option: Optional[TestRunResultDisplay] = TestRunResultDisplay.ALL
    file_output_dir: Optional[str] = None


@dataclass
class CacheConfig:
    write_cache: bool = True
    use_cache: bool = False


@dataclass
class ErrorConfig:
    ignore_errors: bool = False
    skip_on_missing_params: bool = False



from typing import List, Dict, Union
import textwrap
from deepeval.test_case import MLLMImage


class TurnContextualPrecisionTemplate:
    multimodal_rules = """
        --- MULTIMODAL INPUT RULES ---
        - Treat image content as factual evidence.
        - Only reference visual details that are explicitly and clearly visible.
        - Do not infer or guess objects, text, or details not visibly present.
        - If an image is unclear or ambiguous, mark uncertainty explicitly.
        - When evaluating claims, compare them to BOTH textual and visual evidence.
        - If the claim references something not clearly visible, respond with 'idk'.
    """

    @staticmethod
    def generate_verdicts(
        input: str,
        expected_outcome: str,
        retrieval_context: List[str],
        multimodal: bool = False,
    ):
        document_count_str = f" ({len(retrieval_context)} document{'s' if len(retrieval_context) > 1 else ''})"

        # For multimodal, we need to annotate the retrieval context with node IDs
        context_to_display = (
            TurnContextualPrecisionTemplate.id_retrieval_context(
                retrieval_context
            )
            if multimodal
            else retrieval_context
        )

        multimodal_note = (
            " (which can be text or an image)" if multimodal else ""
        )

        prompt_template = textwrap.dedent(
            f"""Given the user message, assistant output, and retrieval context, please generate a list of JSON objects to determine whether each node in the retrieval context was remotely useful in arriving at the assistant output.

            {TurnContextualPrecisionTemplate.multimodal_rules if multimodal else ""}

            **
            IMPORTANT: Please make sure to only return in JSON format, with the 'verdicts' key as a list of JSON. These JSON only contain the `verdict` key that outputs only 'yes' or 'no', and a `reason` key to justify the verdict. In your reason, you should aim to quote parts of the context {multimodal_note}.
            Example Retrieval Context: ["Einstein won the Nobel Prize for his discovery of the photoelectric effect", "He won the Nobel Prize in 1968.", "There was a cat."]
            Example User Message: "Who won the Nobel Prize in 1968 and for what?"
            Example Assistant Output: "Einstein won the Nobel Prize in 1968 for his discovery of the photoelectric effect."

            Example:
            {{
                "verdicts": [
                    {{
                        "reason": "It clearly addresses the question by stating that 'Einstein won the Nobel Prize for his discovery of the photoelectric effect.'",
                        "verdict": "yes"
                    }},
                    {{
                        "reason": "The text verifies that the prize was indeed won in 1968.",
                        "verdict": "yes"
                    }},
                    {{
                        "reason": "'There was a cat' is not at all relevant to the topic of winning a Nobel Prize.",
                        "verdict": "no"
                    }}
                ]  
            }}
            Since you are going to generate a verdict for each context, the number of 'verdicts' SHOULD BE STRICTLY EQUAL to that of the contexts.
            **

            User Message:
            {input}

            Assistant Output:
            {expected_outcome}

            Retrieval Context {document_count_str}:
            {context_to_display}

            JSON:
            """
        )

        return prompt_template

    @staticmethod
    def generate_reason(
        input: str,
        score: float,
        verdicts: List[Dict[str, str]],
        multimodal: bool = False,
    ):
        return textwrap.dedent(
            f"""Given the user message, retrieval contexts, and contextual precision score, provide a CONCISE {'summarize' if multimodal else 'summary'} for the score. Explain why it is not higher, but also why it is at its current score.
            The retrieval contexts is a list of JSON with three keys: `verdict`, `reason` (reason for the verdict) and `node`. `verdict` will be either 'yes' or 'no', which represents whether the corresponding 'node' in the retrieval context is relevant to the user message. 
            Contextual precision represents if the relevant nodes are ranked higher than irrelevant nodes. Also note that retrieval contexts is given IN THE ORDER OF THEIR RANKINGS.

            {TurnContextualPrecisionTemplate.multimodal_rules if multimodal else ""}
            
            **
            IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
            Example JSON:
            {{
                "reason": "The score is <contextual_precision_score> because <your_reason>."
            }}


            DO NOT mention 'verdict' in your reason, but instead phrase it as irrelevant nodes. The term 'verdict' {'are' if multimodal else 'is'} just here for you to understand the broader scope of things.
            Also DO NOT mention there are `reason` fields in the retrieval contexts you are presented with, instead just use the information in the `reason` field.
            In your reason, you MUST USE the `reason`, QUOTES in the 'reason', and the node RANK (starting from 1, eg. first node) to explain why the 'no' verdicts should be ranked lower than the 'yes' verdicts.
            When addressing nodes, make it explicit that {'it is' if multimodal else 'they are'} nodes in {'retrieval context' if multimodal else 'retrieval contexts'}.
            If the score is 1, keep it short and say something positive with an upbeat tone (but don't overdo it{',' if multimodal else ''} otherwise it gets annoying).
            **

            Contextual Precision Score:
            {score}

            User Message:
            {input}

            Retrieval Contexts:
            {verdicts}

            JSON:
            """
        )

    @staticmethod
    def generate_final_reason(
        final_score: float, success: bool, reasons: List[str]
    ):
        return textwrap.dedent(
            f"""You are an AI evaluator producing a single final explanation for the TurnContextualPrecisionMetric result.

            Context:
            This metric evaluates conversational contextual precision by determining whether relevant nodes in retrieval context are ranked higher than irrelevant nodes for each interaction. Each interaction yields a reason indicating why relevant nodes were well-ranked or poorly-ranked. You are given all those reasons.

            Inputs:
            - final_score: the averaged score across all interactions.
            - success: whether the metric passed or failed
            - reasons: a list of textual reasons generated from individual interactions.

            Instructions:
            1. Read all reasons and synthesize them into one unified explanation.
            2. Describe patterns of ranking issues, irrelevant nodes appearing before relevant ones, or well-structured retrieval contexts if present.
            3. Do not repeat every reason; merge them into a concise, coherent narrative.
            4. If the metric failed, state the dominant failure modes. If it passed, state why the retrieval context ranking was effective.
            5. Output a single paragraph with no lists, no bullets, no markup.

            Output:
            A single paragraph explaining the final outcome.

            Here's the inputs:

            Final Score: {final_score}
            
            Reasons: 
            {reasons}

            Success: {success}

            Now give me a final reason that explains why the metric passed or failed. Output ONLY the reason and nothing else.

            The final reason:
            """
        )

    @staticmethod
    def id_retrieval_context(
        retrieval_context: List[str],
    ) -> List[Union[str, MLLMImage]]:
        """
        Annotates retrieval context with node IDs for multimodal processing.

        Args:
            retrieval_context: List of contexts (can be strings or MLLMImages)

        Returns:
            Annotated list with "Node X:" prefixes
        """
        annotated_retrieval_context = []
        for i, context in enumerate(retrieval_context):
            if isinstance(context, str):
                annotated_retrieval_context.append(f"Node {i + 1}: {context}")
            elif isinstance(context, MLLMImage):
                annotated_retrieval_context.append(f"Node {i + 1}:")
                annotated_retrieval_context.append(context)
        return annotated_retrieval_context



import logging
import deepeval

from typing import Optional
from deepeval.telemetry import capture_tracing_integration
from deepeval.tracing.context import current_span_context, current_trace_context
from deepeval.tracing.tracing import Observer
from deepeval.tracing.types import LlmSpan
from deepeval.config.settings import get_settings


logger = logging.getLogger(__name__)


try:
    from crewai.events import BaseEventListener
    from crewai.events import (
        CrewKickoffStartedEvent,
        CrewKickoffCompletedEvent,
        LLMCallStartedEvent,
        LLMCallCompletedEvent,
        AgentExecutionStartedEvent,
        AgentExecutionCompletedEvent,
        ToolUsageStartedEvent,
        ToolUsageFinishedEvent,
        KnowledgeRetrievalStartedEvent,
        KnowledgeRetrievalCompletedEvent,
    )

    crewai_installed = True
except ImportError as e:
    if get_settings().DEEPEVAL_VERBOSE_MODE:
        if isinstance(e, ModuleNotFoundError):
            logger.warning(
                "Optional crewai dependency not installed: %s",
                e.name,
                stacklevel=2,
            )
        else:
            logger.warning(
                "Optional crewai import failed: %s",
                e,
                stacklevel=2,
            )

    crewai_installed = False

IS_WRAPPED_ALL = False


def is_crewai_installed():
    if not crewai_installed:
        raise ImportError(
            "CrewAI is not installed. Please install it with `pip install crewai`."
        )


class CrewAIEventsListener(BaseEventListener):
    def __init__(self):
        is_crewai_installed()
        super().__init__()
        self.span_observers: dict[str, Observer] = {}

    @staticmethod
    def get_tool_execution_id(source, event) -> str:
        source_id = id(source)
        task_id = getattr(event, "task_id", "unknown")
        agent_id = getattr(event, "agent_id", "unknown")
        tool_name = getattr(event, "tool_name", "unknown")
        execution_id = f"tool_{source_id}_{task_id}_{agent_id}_{tool_name}"

        return execution_id

    @staticmethod
    def get_knowledge_execution_id(source, event) -> str:
        source_id = id(source)
        agent_id = id(event.agent) if hasattr(event, "agent") else "unknown"
        execution_id = f"_knowledge_{source_id}_{agent_id}"

        return execution_id

    def setup_listeners(self, crewai_event_bus):
        @crewai_event_bus.on(CrewKickoffStartedEvent)
        def on_crew_started(source, event: CrewKickoffStartedEvent):
            # Assuming that this event is called in the crew.kickoff method
            current_span = current_span_context.get()

            # set the input
            if current_span:
                current_span.input = event.inputs

            # set trace input
            current_trace = current_trace_context.get()
            if current_trace:
                current_trace.input = event.inputs

        @crewai_event_bus.on(CrewKickoffCompletedEvent)
        def on_crew_completed(source, event: CrewKickoffCompletedEvent):
            # Assuming that this event is called in the crew.kickoff method
            current_span = current_span_context.get()

            # set the output
            if current_span:
                current_span.output = str(event.output)

            # set trace output
            current_trace = current_trace_context.get()
            if current_trace:
                current_trace.output = str(event.output)

        @crewai_event_bus.on(LLMCallStartedEvent)
        def on_llm_started(source, event: LLMCallStartedEvent):
            # Assuming that this event is called in the llm.call method
            current_span = current_span_context.get()

            # set the input
            if current_span:
                current_span.input = event.messages

                # set the model
                if isinstance(current_span, LlmSpan):
                    current_span.model = event.model

        @crewai_event_bus.on(LLMCallCompletedEvent)
        def on_llm_completed(source, event: LLMCallCompletedEvent):
            # Assuming that this event is called in the llm.call method
            current_span = current_span_context.get()

            # set the output
            if current_span:
                current_span.output = event.response

        @crewai_event_bus.on(AgentExecutionStartedEvent)
        def on_agent_started(source, event: AgentExecutionStartedEvent):
            # Assuming that this event is called in the agent.execute_task method
            current_span = current_span_context.get()

            # set the input
            if current_span:
                current_span.input = event.task_prompt

        @crewai_event_bus.on(AgentExecutionCompletedEvent)
        def on_agent_completed(source, event: AgentExecutionCompletedEvent):
            # Assuming that this event is called in the agent.execute_task method
            current_span = current_span_context.get()

            # set the output
            if current_span:
                current_span.output = event.output

        @crewai_event_bus.on(ToolUsageStartedEvent)
        def on_tool_started(source, event: ToolUsageStartedEvent):
            observer = Observer(
                span_type="tool",
                func_name=event.tool_name,
                function_kwargs=event.tool_args,
            )
            self.span_observers[self.get_tool_execution_id(source, event)] = (
                observer
            )
            observer.__enter__()

        @crewai_event_bus.on(ToolUsageFinishedEvent)
        def on_tool_completed(source, event: ToolUsageFinishedEvent):
            observer = self.span_observers.pop(
                self.get_tool_execution_id(source, event)
            )
            if observer:
                current_span = current_span_context.get()
                if current_span:
                    current_span.output = event.output
                observer.__exit__(None, None, None)

        @crewai_event_bus.on(KnowledgeRetrievalStartedEvent)
        def on_knowledge_started(source, event: KnowledgeRetrievalStartedEvent):
            observer = Observer(
                span_type="tool",
                func_name="knowledge_retrieval",
                function_kwargs={},
            )
            self.span_observers[
                self.get_knowledge_execution_id(source, event)
            ] = observer
            observer.__enter__()

        @crewai_event_bus.on(KnowledgeRetrievalCompletedEvent)
        def on_knowledge_completed(
            source, event: KnowledgeRetrievalCompletedEvent
        ):
            observer = self.span_observers.pop(
                self.get_knowledge_execution_id(source, event)
            )
            if observer:
                current_span = current_span_context.get()
                if current_span:
                    current_span.input = event.query
                    current_span.output = event.retrieved_knowledge
                observer.__exit__(None, None, None)


def instrument_crewai(api_key: Optional[str] = None):
    is_crewai_installed()
    with capture_tracing_integration("crewai"):
        if api_key:
            deepeval.login(api_key)

        wrap_all()

        CrewAIEventsListener()


def wrap_all():
    global IS_WRAPPED_ALL

    if not IS_WRAPPED_ALL:
        from deepeval.integrations.crewai.wrapper import (
            wrap_crew_kickoff,
            wrap_crew_kickoff_for_each,
            wrap_crew_kickoff_async,
            wrap_crew_kickoff_for_each_async,
            wrap_llm_call,
            wrap_agent_execute_task,
        )

        wrap_crew_kickoff()
        wrap_crew_kickoff_for_each()
        wrap_crew_kickoff_async()
        wrap_crew_kickoff_for_each_async()
        wrap_llm_call()
        wrap_agent_execute_task()

        IS_WRAPPED_ALL = True
